#========================================================================================
# package      DE_EPAGES::Ebay::API::Object::EbayTransaction
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for EbayTransaction.
#========================================================================================
package DE_EPAGES::Ebay::API::Object::EbayTransaction;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::ShopMailType::API::ShopMailType qw( GetShopMailType );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $EbayTransaction->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'EbayTransaction' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $EbayTransaction->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayTransaction = DE_EPAGES::Ebay::API::Object::EbayTransaction->insert($hInfo);
# example      my $Class = LoadClassByAlias('EbayTransaction');
#               my $EbayTransaction = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertEbayTransaction' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $EbayTransaction | new object | object
# hook         OBJ_InsertEbayTransaction | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayTransactionID - object id            - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertEbayTransaction', {'Object' => $self, 'EbayTransactionID' => $self->id });
    delete $hInfo->{$_} foreach qw ( EbayAuction CustomerOrder Customer ExternalTransactionId IsPurchased IsEbayComplete EbayUserName EMail Price Quantity LastMod FeedbackType FeedbackText SubProductAuction ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayTransaction->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteEbayTransaction' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteEbayTransaction | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayTransactionID - object id            - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteEbayTransaction', {'Object' => $self, 'EbayTransactionID' => $self->id });
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $EbayTransaction->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateEbayTransaction' and 'OBJ_AfterUpdateEbayTransaction'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateEbayTransaction | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayTransactionID - object id            - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateEbayTransaction | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>EbayTransactionID - object id            - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateEbayTransaction', {'Object'=>$self, 'Values'=> $hValues, 'EbayTransactionID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateEbayTransaction',  {'Object'=>$self, 'Values'=> $hValues, 'EbayTransactionID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     addToContainer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $Object->addToContainer( $Container, $LanguageID );
#----------------------------------------------------------------------------------------
# description  Adds an LineItemEbayTransaction object to the line item container.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# input        $LanguageID | Language | integer
# input        $SKU | ePages product number (SKU) received from ebay | string
# return       $LineItem | lineitem | object
#========================================================================================
sub addToContainer {
    my $self = shift;
    my ($Container,$LanguageID,$SKU) = @_;

    my $product = $self->get('EbayAuction')->get('Product');

    Error('NoProduct', {'Auction' => $self->get('EbayAuction')}) unless defined $product;

    if(defined $SKU and $product->get('IsMaster')) {
        # This is a multivariation auction
        my $newProduct;
        foreach my $SubAuction ( @{$self->get('EbayAuction')->get('SubAuctions')} ) {
            if (defined $SubAuction->get('Product') ) {
                if( $SubAuction->get('Product')->get('Alias') eq $SKU ){
                    $newProduct = $SubAuction->get('Product');
                    last;
                }
            }
        }
        Error('NoSubProduct', {
            'Auction' => $self->get('EbayAuction'),
            'MasterProduct' => $product->alias,
            'VariationSKU not found' => $SKU
        }) unless defined $newProduct;
        $product = $newProduct;
    }

    my $TaxClass = $product->getWithParent('TaxClass');
    Error('ProductNoTaxClass', {'ProductID'=>$product->id}) unless defined $TaxClass;

    # get line items which are equal to transaction
    my $LineItems = $Container->get('LineItems');
    my @SameTransactionsLineItems = grep { $_->instanceOf('LineItemEbayTransaction') and
                                           defined $_->get('EbayTransaction') and
                                           $_->get('EbayTransaction')->id == $self->id
                                    } @$LineItems;
    return if scalar @SameTransactionsLineItems;

    my $Quantity    = $self->get('Quantity');
    my $Price       = $self->get('Price');

    return $Container->insertLineItem(LoadClassByAlias('LineItemEbayTransaction'), {
        'SKU'               => $product->alias,
        'Name'              => $product->get('NameOrAlias',$LanguageID),
        'TaxClass'          => $TaxClass,
        'Quantity'          => $Quantity,
        'BasePrice'         => $Price / $Quantity,
        'Discount'          => 0,
        'LineItemPrice'     => $Price,
        'EbayTransaction'   => $self,
        'Product'           => $product,
        'TaxRate'           => $Container->taxRate($TaxClass),
        'IsChanged'         => 0,
        'Weight'            => $product->getWithParent('Weight'),
        'WeightUnit'        => $product->getWithParent('WeightUnit'),
        'WeightQuantity'    => $product->getWithParent('PriceQuantity'),
    });
}

#========================================================================================
# function     leaveFeedback
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResult = $EbayTransaction->leaveFeedback( $type, $text );
#----------------------------------------------------------------------------------------
# description  Leave customer feedback on ebay
#----------------------------------------------------------------------------------------
# input        $type | positive (only feedback type possible for merchant) | string
# input        $text | feedback text | string
# return       $hResult | see <function DE_EPAGES::Ebay::API::Object::RemoteEbay::leaveFeedback> | ref.hash
#========================================================================================
sub leaveFeedback {
    my $self = shift;
    my ($type, $text) = @_;

    # check auction
    my $RemoteEbay = DE_EPAGES::Ebay::API::Object::RemoteRESTEbay->new(
        'ShopEnvironment' => $self->get('EbayAuction')->get('EbayShopEnvironment'),
    );

    return $RemoteEbay->leaveFeedback($self, $type, $text);
}

#========================================================================================
# function     sendFulfillmentMail
# state        public
#----------------------------------------------------------------------------------------
# syntax       sendFulfillmentMail( );
#----------------------------------------------------------------------------------------
# description  Send customer mail with instructions on how to complete the transaction
#========================================================================================
sub sendFulfillmentMail {
    my $self = shift;

    my $shop = $self->getSite();
    my $auction = $self->get('EbayAuction');
    my $shopEnv = $auction->get('EbayShopEnvironment');
    my $customer = $self->get('Customer');

    my $MailType = GetShopMailType( $shop, 'FulfillEbayTransaction' ); # always active

    my $name;
    if (defined $customer) {
        my $BillingAddress = $customer->get('BillingAddress');
        $name = $BillingAddress->get('DisplayName') if (defined ($BillingAddress));
    }

    $name = $self->get('EbayUserName') unless (defined $name);

    my %Vars = (
        'Name'                 => $name,
        'EMail'                => $self->get('EMail'),
        'Auction'              => $auction->tleHash,
    );
    my %MailParams = (
        'TO'          => $Vars{'EMail'},
        'HTMLAllowed' => 1,
    );
    my %RegionalParams = (
        Locale      => $shop->get('LocaleID'),
        Language    => lc $shopEnv->get('EbayEnvironment')->get('LanguageID'),
    );
    GetLog->debug('sending fulfill transaction mail ...');
    $MailType->send( $shop, \%MailParams, \%RegionalParams, \%Vars );
}

#========================================================================================
# function     completeOrder
# state        public
#----------------------------------------------------------------------------------------
# syntax       completeOrder( );
#----------------------------------------------------------------------------------------
# description  Set transaction status as complete on ebay
#========================================================================================
sub completeOrder {
    my $self = shift;

    my $Shop = $self->getSite();
    die 'invalid ebay transaction:'.$self->id.' used in Shop:'.$Shop->id if $self->get('EbayAuction')->parent->id != $Shop->child('EbayAuctions')->id;

    if ($self->get('IsPurchased')) {
        GetLog->debug('Ebay Transaction is already purchased.') ;
        return;
    }

    my $EbayAuction =  $self->get('EbayAuction');
    my $Product = $EbayAuction->get('Product');

    if(defined $Product) {
        if ($Product->get('IsMaster')) {
            my $SubAuction;
            $SubAuction = _FindSubAuctionForSKU($EbayAuction, $self->{'Variation'}->{'SKU'});
            Error('NoSubProduct', {
                'Auction' => $EbayAuction,
                'MasterProduct' => $Product->alias,
                'VariationSKU not found' => $self->{'Variation'}->{'SKU'}
            }) unless defined $SubAuction;
        }
    } else {
        Error('NoProduct', {
            'Auction' => $EbayAuction,
            'Product' => $Product->alias
        }) unless defined $Product;
    }
    my $ShopEnv = $EbayAuction->get('EbayShopEnvironment');

    $ShopEnv->remoteEbay()->completeSale(
        $self->get('ExternalTransactionId'),
        $EbayAuction->get('ExternalAuctionId'),
        {
            'Paid'     => 1,
        }
    );
    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Ebay::API::Object::EbayTransaction->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'EbayAuction' => $Parent->parent(),
        'IsPurchased' => 0,
        'IsEbayComplete' => 0,
        'FeedbackType' => 0,
    };
}

1;
