#========================================================================================
# §package      DE_EPAGES::Ebay::API::Object::RemoteEbay
# §state        public
#----------------------------------------------------------------------------------------
# §description  stub object for web service requests to ebay
#========================================================================================
package DE_EPAGES::Ebay::API::Object::RemoteEbay;

##no critic qw(RequireUseStrict)
use if $ENV{PERL_NET_HTTPS_PARALLEL_MODULE} // '' eq 'Net::Curl::Parallel',
    Net::Curl::Parallel;
##use critic

use strict;
use Algorithm::CheckDigits;
use DateTime;
use DateTime::Duration;
use LWP::Parallel::UserAgent;
use Encode qw( decode_utf8 );
use HTTP::Request;
use HTTP::Status qw(HTTP_OK);
use XML::Simple;
use DE_EPAGES::Core::API::Error qw( Error ExistsError GetError );
use DE_EPAGES::Core::API::PerlTools qw( ElementOf GetCurrentDateTime isValid );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );
use DE_EPAGES::Core::API::Log qw( GetLog LogDebug );
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject );
use DE_EPAGES::Object::API::Language qw ( GetPKeyLanguageByCode );
use DE_EPAGES::Ebay::API::Constants;
use DE_EPAGES::Ebay::API::EbayErrors qw (OAuth2TokenErrors TokenOrUserErrors IgnoreWarnings);
use DE_EPAGES::Ebay::API::Table::EbayCall qw (InsertEbayCall);
use DE_EPAGES::Ebay::API::Table::EbayCallType qw (GetPKeyEbayCallTypesByName);
use DE_EPAGES::ThirdPartyConfig::API::ThirdPartyConfig qw (GetDefaultTimeOut);
use DE_EPAGES::Ebay::API::Tools qw (ConvertFromEbayDate);
use DE_EPAGES::Ebay::API::Mail qw (SendTokenErrorMail);
use DE_EPAGES::WebService::API::Client;
use DE_EPAGES::XML::API::DateTimeFormatXML;
use DE_EPAGES::Core::API::Constants qw ( KILO );
use DE_EPAGES::Presentation::API::SiteTLEProcessor qw ( ReplaceTLEs );

use constant ENTRIES_PER_PAGE => 100;
use constant SOAP_ERROR_MESSAGE_LENGTH => 4 * KILO;

#ERRORDEFINITION
sub ErrorDefinition { return {
    'EbayTokenError'            => "The eBay request failed (token problem) for Shop: #ShopAlias (#ShopID) of Store: #Store with Code: #EbayCode.\nReason: #SOAPMessage",
    'EbayNoToken'               => 'eBay request cancelled for Shop: #ShopAlias (#ShopID) of Store: #Store.',
    'NotProviderStore'          => 'The store #Store is not assigned to provider #Provider',
    'EbayRequestNotSuccessful'  => "The eBay request was not successful for Shop: #ShopAlias (#ShopID) of Store: #Store.\nMessage: #Message",
    'EbayRequestFailed'         => "The eBay request failed for Shop: #ShopAlias (#ShopID) of Store: #Store with Code: #EbayCode.\nReason: #SOAPMessage",
    'NoEbayEnvironment'         => 'No eBay Site choosen.',
    'ServerTimeDifferentToEbayTime' => "eBay's time is not synchronous to the server's time (difference: #Minutes min)\nEbayTime  :#EbayTime\nServerTime:#ServerTime",
    'ParameterIncorrect'        => 'Function: #Function called with incorrect Parameter: #Message',
    'EbayNoResponse'            => 'No SOAP response received. Called endpoint #proxy.'
    }
}
#ERRORDEFINITION END

#========================================================================================
sub new {
    my $class = shift;
    my %options = @_;

    if (not defined $options{'Environment'}) {
        Error('NoEbayEnvironment') if (not $options{'ShopEnvironment'});
        my $Env = $options{'ShopEnvironment'}->get('EbayEnvironment');
        Error('NoEbayEnvironment') if (not $Env);
        $options{'Environment'} = $Env;
    }

    $options{'UseProvider'} = 1 unless defined $options{'ShopEnvironment'};

    return bless \%options, $class;
}

#========================================================================================
# §function     store
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $store = $self->store;
#----------------------------------------------------------------------------------------
# §description  Returns the store database object
#----------------------------------------------------------------------------------------
# §return       $Store | Store object | object
#========================================================================================
sub store {
    return LoadRootObject();
}

#========================================================================================
# §function     shopEnvironment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $shop = $self->shopEnvironment;
#----------------------------------------------------------------------------------------
# §description  Returns the shop environment object (for Token purposes)
#----------------------------------------------------------------------------------------
# §return       $shopEnvironment | shop environment object | object
#========================================================================================
sub shopEnvironment {
    return shift->{'ShopEnvironment'};
}

#========================================================================================
# §function     environment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $store = $self->environment;
#----------------------------------------------------------------------------------------
# §description  Returns which eBay Environment should be used
#----------------------------------------------------------------------------------------
# §return       $environment | ebay Environment (incl EbaySiteId) | Object
#========================================================================================
sub environment {
    return shift->{'Environment'};
}

#========================================================================================
# §function     useProvider
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $useProvider = $self->useProvider;
#----------------------------------------------------------------------------------------
# §description  Returns which Token should be used (Provider or Shop)
#----------------------------------------------------------------------------------------
# §return       $useProvider | 1=use Provider Token, 0=use Shop Token | Boolean
#========================================================================================
sub useProvider {
    return shift->{'UseProvider'};
}

#========================================================================================
# §function     getEbayNamespace
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $ebayNamespace = $self->getEbayNamespace;
#----------------------------------------------------------------------------------------
# §description  Returns the namespace that eBay uses in its XML
#----------------------------------------------------------------------------------------
# §return       $ebayNamespace | e.g. "urn:ebay:apis:eBLBaseComponents" | String
#========================================================================================
sub getEbayNamespace {
    return shift->store->get('EbayApiNamespace');
}

#========================================================================================
# §function     variableNamePrefix
# §syntax       $Prefix = $self->variableNamePrefix;
# §state        public
# §description  eBaySandBox or eBay depends on enviroment
# §return       prefix dividing sandbox and live system
#========================================================================================
sub variableNamePrefix {
    return shift->environment->get('IsSandBox') ? 'EbaySandBox' : 'Ebay';
}

#========================================================================================
# §function     _getEbayApiVersion
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $EbayApiVersion = $self->_getEbayApiVersion($MethodName);
#----------------------------------------------------------------------------------------
# §description  Returns the used eBay API version
#               API Version is used from store, if no version is defined
#               for the API call in Constants.pm or no method name is given
#               see DE_EPAGES::Ebay::API::Constants.pm for example
#----------------------------------------------------------------------------------------
# §input        $MethodName (optional)| ebay API method name | string
# §return       $EbayApiVersion | e.g. "707" | String
#========================================================================================
sub _getEbayApiVersion {
    my $self = shift;
    my ($MethodName) = @_;

    return $self->store->get('EbayApiVersion') unless $MethodName;


    # test if we have an other version for this call in Constants.pm
    my $hAPICallVersions = EBAYAPI_CALLVERSIONS();
    return $hAPICallVersions->{$MethodName} if exists $hAPICallVersions->{$MethodName};
    return $self->store->get('EbayApiVersion');

}

#========================================================================================
sub _EbayService {
    my $self = shift;
    my ($methodname) = @_;

    my $proxy = $self->_getEbayProxy($methodname);
    GetLog()->debug('Using eBay Proxy: '.$proxy);

    my $soap = DE_EPAGES::WebService::API::Client->proxy( $proxy, 'timeout' => GetDefaultTimeOut );
    $soap->autotype(0);
    return $soap;
}

#========================================================================================
sub _getEbayProxy {
    my $self = shift;
    my ($methodname) = @_;

    my $endpoint = $self->environment->get('ApiUrl');

    my $siteid = $self->environment->get('EbaySiteId');
    my $appid = $self->store->get($self->variableNamePrefix . 'AppId');

    my $querystring = 'version=' . $self->_getEbayApiVersion($methodname) . '&'.
                      'callname=' . $methodname . '&'.
                      'siteid=' . $siteid . '&' .
                      'appid=' . $appid . '&' .
                      'Routing=default';

    return $endpoint . '?' . $querystring;
}

#========================================================================================
sub _getTokenHeader {
    my $self = shift;

    my $Prefix = $self->variableNamePrefix;
    my $token = $self->useProvider
        ? $self->store->get($Prefix.'ProviderToken')
        : $self->shopEnvironment->get('EbayToken');

    if (not defined $token) {
        Error('EbayNoToken', {
            'Store'       => GetCurrentDBHandle()->cacheKey,
            'Environment' => $self->environment->alias,
            'ShopID'      => $self->useProvider ? ''         : $self->shopEnvironment->getSite()->id,
            'ShopAlias'   => $self->useProvider ? 'Provider' : $self->shopEnvironment->getSite()->alias
        });
    }

    my $header = SOAP::Header->name('RequesterCredentials')->uri($self->getEbayNamespace)
        ->value(\SOAP::Data->value(
            SOAP::Data->name('eBayAuthToken')->value( $token ),
            SOAP::Data->name('Credentials')
                ->value(\SOAP::Data->value(
                    SOAP::Data->name('DevId')->value(    $self->store->get($Prefix.'DevId') ),
                    SOAP::Data->name('AppId')->value(    $self->store->get($Prefix.'AppId') ),
                    SOAP::Data->name('AuthCert')->value( $self->store->get($Prefix.'CertId') ),
                )),
        ))->mustUnderstand(0);

    return $header;
}

#========================================================================================
sub _getLoginHeader {
    my $self = shift;

    my $Prefix = $self->variableNamePrefix;
    my $header = SOAP::Header->name('RequesterCredentials')->uri($self->getEbayNamespace)
        ->value(\SOAP::Data->value(
            SOAP::Data->name('Credentials')
                ->value(\SOAP::Data->value(
                    SOAP::Data->name('DevId')->value( $self->store->get($Prefix.'DevId') ),
                    SOAP::Data->name('AppId')->value( $self->store->get($Prefix.'AppId') ),
                    SOAP::Data->name('AuthCert')->value( $self->store->get($Prefix.'CertId') ),
                    SOAP::Data->name('Username')->value( $self->store->get($Prefix.'ProviderLogin') ),
                    SOAP::Data->name('Password')->value( $self->store->get($Prefix.'ProviderPassword') ),
                )),
        ))->mustUnderstand(0);

    return $header;
}

#========================================================================================
sub _sendRequest { ##no critic qw(ProhibitExcessComplexity)
    my $self = shift;
    my ($methodname, $aRequest, $hReturnFields, $noToken) = @_;
    my $IsLongRunningCall = (shift)?1:0;

    # test if we have a normal Auth'n auth token
    # or if we can use the new REST token
    # abort the request if the token fetch process ends up with an error
    my $OAuthToken = undef;
    my $header;

    eval {
        unless($self->existsRefreshToken() and $self->existsOAuthToken()) {
            $header = ($noToken)?$self->_getLoginHeader():$self->_getTokenHeader();
        } else {
            # if we have no token, we renew the token
            $OAuthToken = $self->oAuthToken();
        }
    };
    return if ExistsError();

    # add the ebay call version for this method
    my $method = SOAP::Data->name( $methodname )->uri($self->getEbayNamespace);
    push @$aRequest, SOAP::Data->name('Version')->value($self->_getEbayApiVersion($methodname));

    # log call
    $self->_logRequest($methodname);

    no warnings 'redefine'; ## no critic qw(ProhibitNoWarnings)
    local *SOAP::Serializer::gen_ns = sub {};
    my $response;
    my $errormsg = undef;
    my $EBayErrorCode = undef;
    my $EBayDetailedMessage = undef;

    for (my $Try = 1; $Try <= MAX_TRIES_FOR_SEND_REQUEST_TOKEN_ERROR; $Try++) {
        my $Error = undef;
        eval {
            my $LogCategory;
            if (LoadRootObject()->get('EbayEnableLogging')) {
                my $ShopID = $self->useProvider
                    ? 'Provider'
                    : $self->shopEnvironment->getSite()->id;
                $LogCategory = "Ebay::RemoteEbay::$ShopID";
            }

            my $soap = $self->_EbayService( $methodname );


            # if we have a rest renew token we test the validity of the auth token
            # and fetch a new one for doing the request
            # use  X-EBAY-API-IAF-TOKEN (rest token in header)
            if($OAuthToken){
                $soap->transport->http_request->headers->push_header('X-EBAY-API-IAF-TOKEN' => $OAuthToken);
            }

            $soap->logCommunication( $LogCategory, sub {
                $response = $soap->call( $method, $header, @$aRequest );
            });
        };
        if (ExistsError()) {
            $Error = GetError();
            $errormsg = $Error->shortMessage;
            my $hFaultDetail = $Error->vars()->{'FaultDetail'};
            if (defined $hFaultDetail and $hFaultDetail =~ /\S/) {
                my $FaultDetail = defined $hFaultDetail->{'FaultDetail'} ? $hFaultDetail->{'FaultDetail'} : undef;
                if (defined $FaultDetail) {
                    $EBayErrorCode = $FaultDetail->{'ErrorCode'};
                    $EBayDetailedMessage = $FaultDetail->{'DetailedMessage'};
                }
            }
        } elsif(not defined $response) {
            Error('EbayNoResponse',{ 'proxy'  => $self->_getEbayProxy($methodname)});
        } elsif($response->fault) {
            $errormsg = $response->faultstring;
        }

        # If the OAuth2 token is invalid we try to renew the token and send the request again
        if ((defined $EBayErrorCode) and ElementOf($EBayErrorCode, OAuth2TokenErrors())) {

            #always write a warning because this error should not happen often
            Warning($Error) if defined $Error;
            $Error = undef;
            # renew the OAuth token and try again
            $self->renewOAuthToken();
            $OAuthToken = $self->oAuthToken();
        } else{
            last; # don't resend
        }

    }# end for

    if (defined $errormsg) {
        $errormsg .= " $EBayDetailedMessage" if defined $EBayDetailedMessage;
        # delete token if token is invalid, see DE_EPAGES::Ebay::API::EbayErrors for further infos.
        my $ErrorCode = 'EbayRequestFailed';
        if (defined $EBayErrorCode and ElementOf($EBayErrorCode, TokenOrUserErrors()))  {
            # delete only old Auth'n auth token
            $self->_deleteInvalidToken() unless $OAuthToken;
            $ErrorCode = 'EbayTokenError';
        }
        my $hError = {
            'SOAPMessage'           => substr($errormsg, 0, SOAP_ERROR_MESSAGE_LENGTH),
            'EnvironmentID'         => $self->environment->id,
            'EnvironmentAlias'      => $self->environment->alias,
            'Store'                 => GetCurrentDBHandle()->cacheKey,
            'ShopEnvironmentID'     => 'Provider',
            'ShopID'                => $self->useProvider ? ''         : $self->shopEnvironment->getSite()->id,
            'ShopAlias'             => $self->useProvider ? 'Provider' : $self->shopEnvironment->getSite()->alias,
            'EbayCode'              => $EBayErrorCode,
        };
        if (not $self->useProvider) {
            my $shopenv = $self->shopEnvironment;
            my $Shop = $shopenv->getSite();
            $hError->{'ShopEnvironmentID'}     = $shopenv->id;
            $hError->{'ShopID'}                = $Shop->id;
            if ($ErrorCode eq 'EbayTokenError') {
                my $System = LoadRootObject();
                my $MailType = $System->folder('MailTypes')->child('ErrorEbaySettings');
                SendTokenErrorMail( $shopenv, $MailType, 'EbayTokenError' ) if $MailType;
            }
        }
        LogDebug('SOAPMessage', $errormsg);
        Error($ErrorCode, $hError);
    }

    my $result ={};
    for my $attribute (qw(Timestamp Ack CorrelationID Errors Version Build ), @{$hReturnFields->{'Fields'}}) {
        $result->{$attribute} = $response->valueof('//'.$attribute);
    };

    foreach my $field (@{$hReturnFields->{'Arrays'}}) {
        my @array = $response->valueof('//'.$field);
        my @resArray = ();
        foreach my $entry (@array) {
            push @resArray, $entry if ref $entry eq 'HASH';
        }
        $result->{$field} = \@resArray;
    }

    # check for problems with server time
    my $serverDT = GetCurrentDateTime();
    my $ebayDT = ConvertFromEbayDate($result->{'Timestamp'});
    my $duration = $ebayDT - $serverDT;


    # ignore long running calls here
    if( (0==$IsLongRunningCall) and (abs $duration->in_units('minutes') > 3) ) {
        my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();

        Error('ServerTimeDifferentToEbayTime', {
            'Minutes'    => $duration->in_units('minutes'),
            'EbayTime'   => $parser->format_datetime($ebayDT),
            'ServerTime' => $parser->format_datetime($serverDT),
        });
    };

    LogDebug('result', $result);

    # we ignore some warnings ( IgnoreWarnings() ) because this are notifications to upgrade to the new ebay API
    if (defined $result->{'Ack'} and $result->{'Ack'} eq 'Warning') {
        my @errors = $response->valueof('//Errors');
        LogDebug('errors', \@errors);

        # add warnings to list of warnings for displaying also if we have a successfull request
        # check if we have warnings and errors...
        # remove warnings listed in IgnoreWarnings()
        my @FilteredWarnings;
        foreach my $error (@errors) {
            if ($error->{'SeverityCode'} eq 'Warning') {
                push @FilteredWarnings, $error if not ElementOf($error->{'ErrorCode'},IgnoreWarnings());
            } else {
                unshift @FilteredWarnings, $error if not ElementOf($error->{'ErrorCode'},IgnoreWarnings());
            }
        }

        Warning('EbayRequestHasWarning', {
            'ErrorCode'  => $result->{'Errors'}->{'ErrorCode'},
            'Message'    => $result->{'Errors'}->{'LongMessage'},
            'EbayErrors' => \@FilteredWarnings
        }) if scalar(@FilteredWarnings);

        # push back ($result->{'Warnings'} was not filtered before)
        $result->{'Warnings'} = \@FilteredWarnings;
        LogDebug('filteredwarnings', $result->{'Warnings'});
    }
    elsif (not defined $result->{'Ack'} or $result->{'Ack'} ne 'Success') {
        my @errors = $response->valueof('//Errors');
        LogDebug('errors', \@errors);

        # check if we have warnings and errors...
        # remove warnings listed in IgnoreWarnings()
        my @FilteredErrors;
        foreach my $error (@errors) {
            # we do not display warnings, if we have errors
            if (not $error->{'SeverityCode'} eq 'Warning') {
                unshift @FilteredErrors, $error if not ElementOf($error->{'ErrorCode'},IgnoreWarnings());
            }
        }

        # push back ($result->{'Errors'} was not filtered before)
        $result->{'Errors'} = \@FilteredErrors;
        Error('EbayRequestNotSuccessful', {
            'Message'    => $result->{'Errors'}->[0]->{'LongMessage'},
            'EbayErrors' => \@FilteredErrors,
            'Store'      => GetCurrentDBHandle()->cacheKey,
            'ShopID'     => $self->useProvider ? ''         : $self->shopEnvironment->getSite()->id,
            'ShopAlias'  => $self->useProvider ? 'Provider' : $self->shopEnvironment->getSite()->alias,
        });
    }

    return $result;
}

#========================================================================================
sub _logRequest {
    my $self = shift;
    my ($methodname) = @_;

    my $calltypes = GetPKeyEbayCallTypesByName( $methodname );

    InsertEbayCall({
        'EbayCallTypeID'    => $calltypes->[0],
        'EbayEnvironmentID' => $self->environment->id,
        'ShopID'            => $self->useProvider?undef:$self->shopEnvironment->getSite()->id,
    }) if ( scalar @$calltypes);

    return 1;
}

#========================================================================================
sub _deleteInvalidToken {
    my $self = shift;

    if ($self->useProvider) {
        # delete token of provider
        GetLog->debug('Removing Invalid Provider Token');
        $self->store->set({$self->variableNamePrefix.'ProviderToken' => undef});
        eval { SynchronizeCache(); };
        Warning( GetError() ) if ExistsError();
    } else {
        # delete token of shop environment
        GetLog->debug('Removing Invalid Token of Shop Environment');
        $self->shopEnvironment->set({
            'EbayToken' => undef,
            'EbayTokenExpiryDate' => undef,
            'EbayUserName' => undef,
        });
    }

    return 1;
}

#========================================================================================
# §function     getCategories
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getCategories($DoCheckOnly, $LevelLimit, $ParentId);
#----------------------------------------------------------------------------------------
# §description  download categories
#----------------------------------------------------------------------------------------
# §input        $DoCheckOnly | flag, 0=only check for new version,
#                                    1=download cat. list | boolean
# §input        $LevelLimit | restrict to tree depth | int
# §input        $ParentId | eBay Category Id, only get childs of this category | int
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>CategoryVersion - version of category list - int
#                   <li>UpdateTime - last update of category list - string
#                   <li>CategoryCount - number of categories - int
#                   <li>Category - array of categories - ref.array
#               </ul> | ref.hash
#========================================================================================
sub getCategories {
    my $self = shift;
    my ($DoCheckOnly, $LevelLimit, $ParentId) = @_;

    my @request = (
                    SOAP::Data->name('CategorySiteID')->value($self->environment->get('EbaySiteId')),
                   );
    push @request, SOAP::Data->name('DetailLevel')->value('ReturnAll')->uri($self->getEbayNamespace) if not $DoCheckOnly;
    push @request, SOAP::Data->name('LevelLimit')->value( $LevelLimit ) if $LevelLimit;
    push @request, SOAP::Data->name('CategoryParent')->value( $ParentId ) if $ParentId;

    my %returnFields = ('Fields' => [qw(CategoryVersion UpdateTime CategoryCount)],
                        'Arrays' => [qw(Category)]);

    my $result = $self->_sendRequest('GetCategories', \@request, \%returnFields,undef,1);

    return $result;
}

#========================================================================================
# §function     getCategory2CS
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getCategory2CS($CategoryID,AttributeSystemVersion);
#----------------------------------------------------------------------------------------
# §description  Retrieves mappings between categories and characteristics sets
#               Retrieves all mappings or just those that match category IDs passed in the call
#               If no CategoryID is given, all mappings are retrieved.
#               If no AttributeSystemVersion is given, all mappings are retrieved,
#               otherwise only mappings with version > AttributeSystemVersion are retrieved
#----------------------------------------------------------------------------------------
# §input        $CategoryID | eBay Category Id (optional) | int
# §input        $AttributeSystemVersion (optional) | A version of the mappings for the site. | int
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>AttributeSystemVersion - Current version of the mappings for the site. - string
#                   <li>Category - array of return url entries - ref.array
#                   <li>SiteWideCharacteristicSets - array of return url entries - ref.array
#               </ul> | ref.hash
#========================================================================================
sub getCategory2CS {
    my $self = shift;
    my ($CategoryID, $AttributeSystemVersion) = @_;

    my @request = ();

    push @request, SOAP::Data->name('AttributeSystemVersion')->value($AttributeSystemVersion) if defined $AttributeSystemVersion;
    push @request, SOAP::Data->name('CategoryID')->value($CategoryID) if defined $CategoryID;
    push @request, SOAP::Data->name('DetailLevel')->value('ReturnAll')->uri($self->getEbayNamespace);

    my %returnFields = ('Fields' => [qw(AttributeSystemVersion)],
                        'Arrays' => [qw(MappedCategoryArray UnmappedCategoryArray SiteWideCharacteristicSets)]);
    my $result = $self->_sendRequest('GetCategory2CS', \@request, \%returnFields,undef,1);

    return $result;
}

#========================================================================================
# §function     getAttributesCS
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getAttributesCS($AttributeSetID,AttributeSystemVersion);
#----------------------------------------------------------------------------------------
# §description  Retrieves attributes for a given AttributeSet
#               If no AttributeSetID is given, all mappings are retrieved. (not usefull)
#               If no AttributeSystemVersion is given, the changed mappings are retrieved,
#               otherwise only mappings with version > AttributeSystemVersion are retrieved
#----------------------------------------------------------------------------------------
# §input        $AttributeSetID | eBay AttributeSetID Id (optional) | int or ref.array
# §input        $AttributeSystemVersion (optional) | A version of the mappings for the site. | int
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>AttributeSystemVersion - Current version of the mappings for the site. - string
#                   <li>AttributeData - xml attribute data string - string
#               </ul> | ref.hash
#========================================================================================
sub getAttributesCS {
    my $self = shift;
    my ($AttributeSetID, $AttributeSystemVersion) = @_;

    my @request = ();

    push @request, SOAP::Data->name('AttributeSystemVersion')->value($AttributeSystemVersion) if defined $AttributeSystemVersion;
    push @request, SOAP::Data->name('AttributeSetID')->value($AttributeSetID) if defined $AttributeSetID;
    push @request, SOAP::Data->name('DetailLevel')->value('ReturnAll')->uri($self->getEbayNamespace);

    my %returnFields = ('Fields' => [qw(AttributeSystemVersion AttributeData )],
                        'Arrays' => []);
    my $result = $self->_sendRequest('GetAttributesCS', \@request, \%returnFields,undef,1);

    return $result;
}



#========================================================================================
# §function     verifyAddItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->verifyAddItem($Auction);
#----------------------------------------------------------------------------------------
# §description  upload auction for test purposes
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | EbayAuction object
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>ItemID - ebay item id of auction - string
#                   <li>Fee - array of fees - ref.array
#               </ul> | ref.hash
#========================================================================================
sub verifyAddItem {
    my $self = shift;
    my ($Auction) = @_;

    my $aItemValues = $self->_getEbayItemFromAuction( $Auction );

    my @request = (
                    SOAP::Data->name('Item')->value(
                        \SOAP::Data->value(@$aItemValues),
                    )->uri($self->getEbayNamespace),
                   );

    my %returnFields = ('Fields' => [qw(ItemID)],
                        'Arrays' => [qw(Fee)]);

    my $result = $self->_sendRequest('VerifyAddItem', \@request, \%returnFields);

    return $result;
}


#========================================================================================
# §function     verifyAddFixedPriceItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->verifyAddFixedPriceItem($Auction);
#----------------------------------------------------------------------------------------
# §description  upload fixed price and multi variation auctions for test purposes
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | EbayAuction object
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>ItemID - ebay item id of auction - string
#                   <li>Fee - array of fees - ref.array
#               </ul> | ref.hash
#========================================================================================
sub verifyAddFixedPriceItem {
    my $self = shift;
    my ($Auction) = @_;

    my $aItemValues = $self->_getEbayItemFromAuction( $Auction );

    my @request = (
                    SOAP::Data->name('Item')->value(
                        \SOAP::Data->value(@$aItemValues),
                    )->uri($self->getEbayNamespace),
                   );

    my %returnFields = ('Fields' => [qw(ItemID)],
                        'Arrays' => [qw(Fee)]);

    my $result = $self->_sendRequest('VerifyAddFixedPriceItem', \@request, \%returnFields);

    return $result;
}


#========================================================================================
# §function     addItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->addItem($Auction);
#----------------------------------------------------------------------------------------
# §description  upload auction to ebay
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | EbayAuction object
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>ItemID - ebay item id of auction - string
#                   <li>StartTime - start of auction - string
#                   <li>EndTime - end of auction - string
#                   <li>Fee - array of fees - ref.array
#               </ul> | ref.hash
#========================================================================================
sub addItem {
    my $self = shift;
    my ($Auction) = @_;

    my $aItemValues = $self->_getEbayItemFromAuction( $Auction );

    my @request = (
                    SOAP::Data->name('Item')->value(
                        \SOAP::Data->value(@$aItemValues),
                    )->uri($self->getEbayNamespace),
                   );

    my %returnFields = ('Fields' => [qw(ItemID StartTime EndTime)],
                        'Arrays' => [qw(Fee)]);

    my $result = $self->_sendRequest('AddItem', \@request, \%returnFields);

    return $result;
}


#========================================================================================
# §function     addFixedPriceItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->addFixedPriceItem($Auction);
#----------------------------------------------------------------------------------------
# §description  upload auction to ebay
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | EbayAuction object
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#                   <li>ItemID - ebay item id of auction - string
#                   <li>StartTime - start of auction - string
#                   <li>EndTime - end of auction - string
#                   <li>Fee - array of fees - ref.array
#               </ul> | ref.hash
#========================================================================================
sub addFixedPriceItem {
    my $self = shift;
    my ($Auction) = @_;

    my $aItemValues = $self->_getEbayItemFromAuction( $Auction );

    my @request = (
                    SOAP::Data->name('Item')->value(
                        \SOAP::Data->value(@$aItemValues),
                    )->uri($self->getEbayNamespace),
                   );

    my %returnFields = ('Fields' => [qw(ItemID StartTime EndTime)],
                        'Arrays' => [qw(Fee)]);

    my $result = $self->_sendRequest('AddFixedPriceItem', \@request, \%returnFields);

    return $result;
}


#========================================================================================
# §function     getSellerList
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getSellerList($startTimeFrom, $page);
#----------------------------------------------------------------------------------------
# §description  Returns a list of items a seller has listed for auction.
#----------------------------------------------------------------------------------------
# §input        $dateFrom | start with this date | DateTime object
# §return       $Items | array of items | ref.array.hash
#========================================================================================
sub getSellerList {
    my $self = shift;
    my ($dateFrom) = @_;

    my $now = DateTime->now;
    my $maxTimeRange = DateTime::Duration->new(days => 119);
    if (!$dateFrom) {
        $dateFrom = $now - $maxTimeRange;
    } else {
        # 0 is == / -1 if From < To / 1 if From > To
        if (1 == DateTime->compare($dateFrom, $now)) {
            Warning('ParameterIncorrect', {'Function' => 'getSellerList' ,'Message' => 'dateFrom is in the future'});
            $dateFrom = $now - $maxTimeRange;
        }
    }

    my @Items = ();
    while ($dateFrom < $now) {
        my $dateTo = $dateFrom->clone;
        $dateTo->add($maxTimeRange);
        $dateTo = $now if ($dateTo > $now);

        my $result;
        # with https_proxy, parallel works only with 'Net::SSL'
        my $CanParallel = $ENV{https_proxy} ? 0 : 1;
        $CanParallel = 1 if $ENV{PERL_NET_HTTPS_SSL_SOCKET_CLASS} eq 'Net::SSL';
        $CanParallel = 1 if $ENV{PERL_NET_HTTPS_PARALLEL_MODULE} eq 'Net::Curl::Parallel';
        # can we use parallel requests for getSellerList?
        if ( $CanParallel and REMOTE_API_PARALLEL_REQUESTS >= 1) {
            # fetch 1st page to get the number of all items for parallelizing the requests
            $result = $self->_getSellerList($dateFrom->clone, $dateTo->clone, 1);
            if (defined $result and (ref $result eq 'HASH') ){
                push(@Items, @{$result->{'Item'}});
                # do we have to fetch more pages with items?
                if ($result->{'HasMoreItems'} eq 'true') {
                    my $CountOfPages = $result->{'PaginationResult'}->{'TotalNumberOfPages'};
                    # get the TotalNumberOfPages with parallel requests to the eBay API, starting with page 2 (the 1st we have already fetched)
                    $self->_getSellerListParallel(\@Items,$dateFrom->clone, $dateTo->clone,$CountOfPages, 2);
                }
            }
            $dateFrom = $dateTo;

        } else {
            # old handling
            my $page = 1;
            do {
                $result = $self->_getSellerList($dateFrom->clone, $dateTo->clone, $page++);
                 push(@Items, @{$result->{'Item'}});
            }
            while($result->{'HasMoreItems'} eq 'true');
            $dateFrom = $dateTo;
        }
    }

    return \@Items;
}

#========================================================================================
# §function     _getSellerListParallel
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_getSellerListParallel($Items,$dateFrom, $dateTo, $CountOfPages, $StartPage);
#----------------------------------------------------------------------------------------
# §description  get items of seller
#----------------------------------------------------------------------------------------
# §input        $Items | items are stored here | array of hashes
# §input        $dateFrom | start with this date | DateTime object
# §input        $dateTo | enddate | DateTime object
# §input        $CountOfPages | number of pages to fetch from the eBay API | integer
# §input        $StartPage | start with this page number, default: 1 | integer
# §return       $result | result of webservice call | ref.hash
#========================================================================================
sub _getSellerListParallel {
    my $self = shift;
    my ($Items, $dateFrom, $dateTo, $CountOfPages, $StartPage) = @_;
    $StartPage = 1 unless $StartPage;


    #convert date for request
    my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();
    # for ebay we need the UTC time
    $dateFrom->set_time_zone('UTC');
    $dateTo->set_time_zone('UTC');


    # we have here an own request handling without using soap by our webservice modules
    # prepare request header for an XML API call "GetSellerList" (yes, we do not use soap here)
    my $aRequests = [];

    my $APIUrl = $self->environment->get('IsSandBox') ? REMOTE_GETSELLERLIST_XML_SANDBOX_API_URL : REMOTE_GETSELLERLIST_XML_API_URL;
    my $Prefix = $self->variableNamePrefix;
    my $hHeader = { 'X-EBAY-API-DEV-NAME' => $self->store->get($Prefix.'DevId'),
                    'X-EBAY-API-APP-NAME' => $self->store->get($Prefix.'AppId'),
                    'X-EBAY-API-CERT-NAME' => $self->store->get($Prefix.'CertId'),
                    'X-EBAY-API-COMPATIBILITY-LEVEL' => $self->_getEbayApiVersion('GetSellerList'),
                    'X-EBAY-API-SITEID' => $self->environment->get('EbaySiteId'),
                    'X-EBAY-API-CALL-NAME' => 'GetSellerList',
                    'X-EBAY-API-REQUEST-ENCODING' => 'XML',
                    'Content-Type' => 'text/xml',
                  };

    # test if we have a normal Auth'n auth token
    # or if we can use the new REST token
    my $Token;
    unless($self->existsRefreshToken() and $self->existsOAuthToken()) {
        $Token  = $self->shopEnvironment->get('EbayToken');
    } else {
        # add the OAuth token if needed
        $hHeader->{'X-EBAY-API-IAF-TOKEN'} = $self->oAuthToken();
    }

    foreach my $pagenr ($StartPage..$CountOfPages) {
        # decide if we have to use the sandbox or live system
        my $XMLRequest = HTTP::Request->new(POST => $APIUrl);

        # add header with credentials
        $XMLRequest->header(%$hHeader);

        my $xml = XML::Simple->new();
        my $XMLTree = {
           'xmlns'       => 'urn:ebay:apis:eBLBaseComponents',
           'StartTimeFrom'  => [$parser->format_datetime($dateFrom)],
           'StartTimeTo'    => [$parser->format_datetime($dateTo)],
           'DetailLevel'    => ['ItemReturnDescription'],
           'Pagination'     => { 'ComplexType'    => 'PaginationType',
                                 'EntriesPerPage' => [REMOTE_GETSELLERLIST_ENTRIESPERPAGE],
                                 'PageNumber'     => [$pagenr] },
           'OutputSelector' => ['HasMoreItems,ItemArray.Item.ItemID,ItemArray.Item.ListingDetails.EndTime,ItemArray.Item.SellingStatus.BidCount,'.
                                'ItemArray.Item.SellingStatus.CurrentPrice,ItemArray.Item.SellingStatus.ListingStatus,ItemArray.Item.SellingStatus.QuantitySold,'.
                                'PageNumber,PaginationResult,ReturnedItemCountActual'],
        };

        # use the old token if we have no OAuth token
        $XMLTree->{'RequesterCredentials'} = {'eBayAuthToken' => [$Token] } if defined $Token;

        my $dec = '<?xml version="1.0" encoding="utf-8"?>';
        my $XmlContent = $xml->XMLout($XMLTree, XMLDecl => $dec, RootName => 'GetSellerListRequest');

        $XMLRequest->add_content($XmlContent);
        LogDebug('XMLRequestContent',$XmlContent);

        # collect requests for parallel handling
        push @$aRequests, $XMLRequest;
    }

    # send request in parallel
    my $Responses = $self->_sendParallelRequest($aRequests);

    foreach my $response (@$Responses) {
        LogDebug('Response', $response);

        # test response code and set errors
        if ($response->{code} != HTTP_OK) {
            Error('EbayGetSellerListFailed',{ 'url'   => $response->{uri},
                                              'code'  => $response->{code},
                                              'message' => $response->{message},
                  });
        }

        # decode XML to perl structure
        my $NoXML;
        my $xml = XML::Simple->new();

        #printf "Response %s\n",$response->{content};

        my $Decoded = decode_utf8($response->{content});
        $NoXML = eval { $xml->XMLin( $Decoded ); };

        #LogDebug('ResponseContent',$NoXML);


        if (ExistsError()) {
            my $Error = GetError();
            # collect all errors, but let the script continue
            Warning('XMLinError',{'XML' => $Decoded,'Error' => $Error->message});

            #go to next response and continue
            next;
        }

        if (defined $NoXML->{'Ack'} and $NoXML->{'Ack'} eq 'Warning') {
            LogDebug('Warnings', $NoXML);
            my $hErrors = $NoXML->{'Errors'};
            Warning('EbayRequestHasWarning', {
                    'ErrorCode'  => $hErrors->{'ErrorCode'},
                    'Message'    => $hErrors->{'LongMessage'},
                    'EbayErrors' => $hErrors
                   });
        }
        elsif (not defined $NoXML->{'Ack'} or $NoXML->{'Ack'} ne 'Success') {
            my $hErrors = $NoXML->{'Errors'};
            Error('EbayRequestNotSuccessful', {
                'Message'    => $hErrors->{'LongMessage'},
                'EbayErrors' => $hErrors,
                'Store'      => GetCurrentDBHandle()->cacheKey,
                'ShopID'     => $self->useProvider ? ''         : $self->shopEnvironment->getSite()->id,
                'ShopAlias'  => $self->useProvider ? 'Provider' : $self->shopEnvironment->getSite()->alias,
            });
        };

        # rework current price entries and add elements to array
        # sometimes "CurrentPrice" entry is a HASH and needs to reassigned
        my $Item = $NoXML->{'ItemArray'}->{'Item'};

        if( ref $Item eq 'HASH' ) {
            if  (ref $Item->{'SellingStatus'}->{'CurrentPrice'} eq 'HASH') {
                $Item->{'SellingStatus'}->{'CurrentPrice'} = $Item->{'SellingStatus'}->{'CurrentPrice'}->{'content'};
            }
            push @$Items, $Item;
        }elsif( ref $Item eq 'ARRAY') {
            foreach my $Elem (@{$Item}) {
                if  (ref $Elem->{'SellingStatus'}->{'CurrentPrice'} eq 'HASH') {
                    $Elem->{'SellingStatus'}->{'CurrentPrice'} = $Elem->{'SellingStatus'}->{'CurrentPrice'}->{'content'};
                }
                push @$Items,$Elem;
            }
        }
    }

    return;

}

sub _sendParallelRequest {
    my $self = shift;
    my ($aRequests) = @_;

    # use the UA for handling parallel requests

    if ($ENV{PERL_NET_HTTPS_PARALLEL_MODULE} eq 'Net::Curl::Parallel') {
        # create a parallel user agent
        # REMOTE_GETSELLERLIST_UA_TIMEOUT is in seconds
        # wheras connect_timeout and request_timeout are in milliseconds
        my $pua = Net::Curl::Parallel->new(
            agent           => 'Net::Curl::Parallel/v0.1',
            slots           => REMOTE_API_PARALLEL_REQUESTS,
            max_redirects   => 0,
            connect_timeout => REMOTE_GETSELLERLIST_UA_TIMEOUT * KILO,
            request_timeout => REMOTE_GETSELLERLIST_UA_TIMEOUT * KILO,
            );

        # register requests for parallel handling
        my @Responses;
        foreach my $Request (@$aRequests) {
            #print "Registering '".$Request->uri->as_string."'\n";
            LogDebug('RegisterURL',$Request->uri);
            push @Responses, { uri => $Request->uri->as_string };
            $pua->try($Request);
        }

        # start and wait for requests, then return results
        my @performed = $pua->perform;

        my $count = 0;
        foreach my $response (@performed) {
            ${Responses[$count]}{code} = $response->status();
            ${Responses[$count]}{message} = $response->error();
            ${Responses[$count]}{content} = $response->content();
            $count++;
        }
        return \@Responses;

    } else {
        # create a parallel user agent
        my $pua = LWP::Parallel::UserAgent->new();
        $pua->in_order(0);  # handle requests in order of registration
        $pua->redirect(0);  # follow redirects
        $pua->timeout(REMOTE_GETSELLERLIST_UA_TIMEOUT);  # in seconds
        # max number of requests to process simultaneously:
        $pua->max_req(REMOTE_API_PARALLEL_REQUESTS);

        # register requests for parallel handling
        foreach my $Request (@$aRequests) {
            #print "Registering '".$Request->uri->as_string."'\n";
            LogDebug('RegisterURL',$Request->uri);
            if ( my $res = $pua->register ($Request) ) {
                Error('ErrorRegisteringRequest', {'URL' => $Request->uri->as_string, 'error' => $res->error_as_HTML });
            }
        }

        # start and wait for requests, then return results
        my $performed = $pua->wait(REMOTE_GETSELLERLIST_UA_TIMEOUT);

        my @Responses;
        foreach my $response (keys %$performed) {
            push @Responses, {
                uri => $performed->{$response}->response->request->uri->as_string,
                code => $performed->{$response}->response->code,
                message => $performed->{$response}->response->message,
                content => $performed->{$response}->response->content,
            }
        }
        return \@Responses;
    }

}

#========================================================================================
# §function     _getSellerList
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_getSellerList($dateFrom, $dateTo, 2);
#----------------------------------------------------------------------------------------
# §description  get items of seller
#----------------------------------------------------------------------------------------
# §input        $dateFrom | start with this date | DateTime object
# §input        $dateTo | enddate | DateTime object
# §input        $page | page, default: 1 | integer
# §return       $result | result of webservice call | ref.hash
#========================================================================================
sub _getSellerList {
    my $self = shift;
    my ($dateFrom, $dateTo, $page) = @_;
    $page = 1 unless $page;

    # for ebay we need the UTC time
    $dateFrom->set_time_zone('UTC');
    $dateTo->set_time_zone('UTC');

    my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();
    my @request = (
        SOAP::Data->name('StartTimeFrom')->value($parser->format_datetime($dateFrom)),
        SOAP::Data->name('StartTimeTo')->value($parser->format_datetime($dateTo)),
        SOAP::Data->name('Pagination')->value({'EntriesPerPage' => REMOTE_GETSELLERLIST_ENTRIESPERPAGE, 'PageNumber' => $page})->uri($self->getEbayNamespace),
        SOAP::Data->name('DetailLevel')->value('ItemReturnDescription')->uri($self->getEbayNamespace),
        SOAP::Data->name('OutputSelector')->value(join(',',REMOTE_GETSELLERLIST_OUTPUTSELECTOR)),
    );

    my %returnFields = ('Fields' => [qw(HasMoreItems PaginationResult PageNumber ReturnedItemCountActual)],
                        'Arrays' => [qw(Item)]);

    return $self->_sendRequest('GetSellerList', \@request, \%returnFields);
}

#========================================================================================
# §function     getItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getItem($externalAuctionId);
#----------------------------------------------------------------------------------------
# §description  get a single item
#----------------------------------------------------------------------------------------
# §input        $externalAuctionId | auction id of item | string
# §return       $hItem | info of item | ref.hash
#========================================================================================
sub getItem {
    my $self = shift;
    my ($externalAuctionId) = @_;

    my @request = (
        SOAP::Data->name('ItemID')->value($externalAuctionId),
        SOAP::Data->name('DetailLevel')->value('ItemReturnDescription')->uri($self->getEbayNamespace)
    );

    my %returnFields = ('Fields' => [qw(Item)],
                        'Arrays' => []);

    my $result = $self->_sendRequest('GetItem', \@request, \%returnFields);

    return $result->{'Item'};
}

#========================================================================================
# §function     getSellerTransactions
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getSellerTransactions($modTimeFrom, $page);
#----------------------------------------------------------------------------------------
# §description  get transactions of seller
#----------------------------------------------------------------------------------------
# §input        $modTimeFrom | start with this date | DateTime object
# §return       $Transactions | array of transactions | ref.array.hash
#========================================================================================
sub getSellerTransactions {
    my $self = shift;
    my ($modTimeFrom) = @_;

    my $now = DateTime->now;
    my $dur30days = DateTime::Duration->new(days => 30);
    if (!$modTimeFrom) {
        $modTimeFrom = $now - $dur30days;
    }

    my @Transactions = ();
    while ($modTimeFrom < $now) {
        my $modTimeTo = $modTimeFrom->clone;
        $modTimeTo->add($dur30days);
        $modTimeTo = $now if ($modTimeTo > $now);
        my $page = 1;
        my $result;
        do {
            $result = $self->_getSellerTransactions($modTimeFrom->clone, $modTimeTo->clone, $page++);
            push(@Transactions, @{$result->{'Transaction'}}) if defined $result && defined $result->{'Transaction'};
        }
        while($result->{'HasMoreTransactions'} eq 'true');
        $modTimeFrom = $modTimeTo;
    }

    return \@Transactions;
}

#========================================================================================
# §function     _getSellerTransactions
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_getSellerTransactions($modTimeFrom, $page);
#----------------------------------------------------------------------------------------
# §description  get transactions of item
#----------------------------------------------------------------------------------------
# §input        $modTimeFrom | start with this date | DateTime object
# §input        $modTimeTo | end date | DateTime object
# §input        $page | page, default: 1 | integer
# §return       $result | result of webservice call | ref.hash
#========================================================================================
sub _getSellerTransactions {
    my $self = shift;
    my ($modTimeFrom, $modTimeTo, $page) = @_;
    $page = 1 unless $page;

    # for ebay we need the UTC time
    $modTimeFrom->set_time_zone('UTC');
    $modTimeTo->set_time_zone('UTC');

    my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();
    my @request = (
        SOAP::Data->name('ModTimeFrom')->value($parser->format_datetime($modTimeFrom)),
        SOAP::Data->name('ModTimeTo')->value($parser->format_datetime($modTimeTo)),
        SOAP::Data->name('Pagination')->value({'EntriesPerPage' => ENTRIES_PER_PAGE, 'PageNumber' => $page})->uri($self->getEbayNamespace),
    );

    my %returnFields = ('Fields' => [qw(HasMoreTransactions PaginationResult PageNumber ReturnedTransactionCountActual)],
                        'Arrays' => [qw(Transaction)]);

    return $self->_sendRequest('GetSellerTransactions', \@request, \%returnFields);
}

#========================================================================================
# §function     getItemTransactions
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getItemTransactions($externalAuctionId, $modTimeFrom);
#----------------------------------------------------------------------------------------
# §description  get transactions of item
#----------------------------------------------------------------------------------------
# §input        $externalAuctionId | auction id of item | string
# §input        $modTimeFrom | start with this date | DateTime object
# §return       $Transactions | array of transactions | ref.array.hash
#========================================================================================
sub getItemTransactions {
    my $self = shift;
    my ($externalAuctionId, $modTimeFrom) = @_;

    my $now = DateTime->now;
    my $dur30days = DateTime::Duration->new(days => 30);
    if (!$modTimeFrom) {
        $modTimeFrom = $now - $dur30days;
    }

    my @Transactions = ();
    while ($modTimeFrom < $now) {
        my $modTimeTo = $modTimeFrom->clone;
        $modTimeTo->add($dur30days);
        $modTimeTo = $now if ($modTimeTo > $now);
        my $result = $self->_getItemTransactions($externalAuctionId, $modTimeFrom->clone, $modTimeTo->clone);
        push(@Transactions, @{$result->{'Transaction'}});
        $modTimeFrom = $modTimeTo;
    }

    return \@Transactions;
};

#========================================================================================
# §function     _getItemTransactions
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->_getItemTransactions($externalAuctionId, $modTimeFrom, $modTimeTo);
#----------------------------------------------------------------------------------------
# §description  get transactions of item
#----------------------------------------------------------------------------------------
# §input        $externalAuctionId | auction id of item | string
# §input        $modTimeFrom | start with this date | DateTime object
# §input        $modTimeTo | end date | DateTime object
# §return       $result | result of webservice call | ref.hash
#========================================================================================
sub _getItemTransactions {
    my $self = shift;
    my ($externalAuctionId, $modTimeFrom, $modTimeTo) = @_;

    # for ebay we need the UTC time
    $modTimeFrom->set_time_zone('UTC');
    $modTimeTo->set_time_zone('UTC');

    my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();
    my @request = (
        SOAP::Data->name('ItemID')->value($externalAuctionId),
        SOAP::Data->name('ModTimeFrom')->value($parser->format_datetime($modTimeFrom)),
        SOAP::Data->name('ModTimeTo')->value($parser->format_datetime($modTimeTo)),
    );

    my %returnFields = ('Fields' => [qw(HasMoreTransactions PaginationResult PageNumber ReturnedTransactionCountActual)],
                        'Arrays' => [qw(Transaction)]);

    return $self->_sendRequest('GetItemTransactions', \@request, \%returnFields);
}

#========================================================================================
# §function     leaveFeedback
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->leaveFeedback($Transaction, $CommentType, $CommentText);
#----------------------------------------------------------------------------------------
# §description  leave feedback for a transaction
#----------------------------------------------------------------------------------------
# §input        $Transaction | ebay transaction | EbayTransaction object
# §input        $CommentType | comment type | string
# §input        $CommentText | comment text| string
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#               </ul> | ref.hash
#========================================================================================
sub leaveFeedback {
    my $self = shift;
    my ($Transaction, $CommentType, $CommentText) = @_;

    my @request = (
        SOAP::Data->name('TargetUser')->value( $Transaction->get('EbayUserName') ),
        SOAP::Data->name('ItemID')->value( $Transaction->get('EbayAuction')->get('ExternalAuctionId') ),
        SOAP::Data->name('TransactionID')->value( $Transaction->get('ExternalTransactionId') ),
        SOAP::Data->name('CommentType')->value( $CommentType ),
        SOAP::Data->name('CommentText')->value( $CommentText ),
    );

    my %returnFields = ('Fields' => [qw()],
                        'Arrays' => [qw()]);

    my $result = $self->_sendRequest('LeaveFeedback', \@request, \%returnFields);

    return $result;
}

#========================================================================================
# §function     getStore
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getStore($Transaction, $CommentType, $CommentText);
#----------------------------------------------------------------------------------------
# §description  retrieves the store of a ebay seller
#----------------------------------------------------------------------------------------
# §return       $result | with following keys:
#               <ul>
#                   <li>CustomCategories - list of store categories - ref.array.hash
#                   <li>Description - store description - string
#                   <li>Name - store name - string
#                   <li>SubscriptionLevel - store's level - string
#                   <li>URL - store's url - string
#                   <li>and some more info
#               </ul> | ref.hash
#========================================================================================
sub getStore {
    my $self = shift;

    my @request = ();

    my %returnFields = ('Fields' => [qw(Store)],
        # "Arrays" only works for first-level xml elements. So "ChildCategory"
        # is non-functional here.
                        'Arrays' => [qw( CustomCategory )]);

    my $result = $self->_sendRequest('GetStore', \@request, \%returnFields);

    my @LeafCategories = ();
    LogDebug( 'Response', $result);
    $self->_parseStoreCategories('',\@LeafCategories,$result->{'CustomCategory'});
    $result->{'Store'}->{'CustomCategories'} = \@LeafCategories;

    return $result->{'Store'};
}

#========================================================================================
# §function     _parseStoreCategories
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _parseStoreCategories( $Path, $LeafCategories, $StoreCategories );
#----------------------------------------------------------------------------------------
# §description  parses store category tree recursively to display in a list.
#----------------------------------------------------------------------------------------
# §input        $Path | current category path  | string
# §input        $LeafCategories | built-up list of categories | string.array.ref
# §input        $StoreCategories | store category list to parse | hash.array.ref
# §return       $LeafCategories | extended list of categories | string.array.ref
#========================================================================================
sub _parseStoreCategories {
    my $self = shift;
    my ($Path, $LeafCategories, $StoreCategories) = @_;
    foreach my $hCategory (@$StoreCategories) {
        if (exists $hCategory->{'ChildCategory'}) {
            if (ref $hCategory->{'ChildCategory'} eq 'HASH') {
                $self->_parseStoreCategories($Path.$hCategory->{'Name'}.' / ', $LeafCategories, [$hCategory->{'ChildCategory'}]);
            } elsif (ref $hCategory->{'ChildCategory'} eq 'ARRAY') {
                $self->_parseStoreCategories($Path.$hCategory->{'Name'}.' / ', $LeafCategories, $hCategory->{'ChildCategory'});
            }
        } else {
            $hCategory->{'Path'} = $Path;
            push @$LeafCategories, $hCategory;
        }
    }
    return $LeafCategories;
}

#========================================================================================
# §function     getAllBidders
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getAllBidders($externalAuctionId);
#----------------------------------------------------------------------------------------
# §description  get a all bidders from auction, for second offer change
#----------------------------------------------------------------------------------------
# §input        $externalAuctionId | auction id of item | string
# §return       $hBidders | info of item | ref.hash
#========================================================================================
sub getAllBidders {
    my $self = shift;
    my ($externalAuctionId) = @_;

    my @request = (
        SOAP::Data->name('ItemID')->value($externalAuctionId),
        SOAP::Data->name('CallMode')->value('ViewAll')->uri($self->getEbayNamespace)
    );

    my %returnFields = ('Fields' => [],
                        'Arrays' => [qw(Offer)]);

    my $result = $self->_sendRequest('GetAllBidders', \@request, \%returnFields);

    return $result;
}

#========================================================================================
# §function     endItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->endItem($externalAuctionId);
#----------------------------------------------------------------------------------------
# §description  ends an ebay item
#----------------------------------------------------------------------------------------
# §input        $externalAuctionId | auction id of item | string
# §input        $EndingReason | ending reason of item | string (optional)
# §return       $Success | 0 on failure 1 on success.
#========================================================================================
sub endItem {
    my $self = shift;
    my ($externalAuctionId, $EndingReason) = @_;

    $EndingReason //= 'OtherListingError';

    my @request = (
        SOAP::Data->name('ItemID')->value($externalAuctionId),
        SOAP::Data->name('EndingReason')->value($EndingReason),
    );

    my $result = $self->_sendRequest('EndItem', \@request);

    return $result;
}


#========================================================================================
# §function     relistFixedPriceItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->relistFixedPriceItem($Auction);
#----------------------------------------------------------------------------------------
# §description  relists an ebay fixed price item
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | EbayAuction object
# §return       $Success | 0 on failure 1 on success.
#========================================================================================
sub relistFixedPriceItem {
    my $self = shift;
    my ($Auction) = @_;

    my $aItemValues = $self->_getEbayItemFromAuction( $Auction, 1 );

    my @request = (
        SOAP::Data->name('Item')->value(
            \SOAP::Data->value(@$aItemValues),
        )->uri($self->getEbayNamespace),
    );

    my %returnFields = ('Fields' => [qw(ItemID StartTime EndTime)],
                        'Arrays' => [qw(Fee)]);

    my $result = $self->_sendRequest('RelistFixedPriceItem', \@request, \%returnFields);

    return $result;
}

#========================================================================================
# §function     relistItem
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->relistItem($Auction);
#----------------------------------------------------------------------------------------
# §description  relists an ebay auction item
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | EbayAuction object
# §return       $Success | 0 on failure 1 on success.
#========================================================================================
sub relistItem {
    my $self = shift;
    my ($Auction) = @_;

    my $aItemValues = $self->_getEbayItemFromAuction( $Auction, 1 );

    my @request = (
        SOAP::Data->name('Item')->value(
            \SOAP::Data->value(@$aItemValues),
        )->uri($self->getEbayNamespace),
    );

    my %returnFields = ('Fields' => [qw(ItemID StartTime EndTime)],
                        'Arrays' => [qw(Fee)]);

    my $result = $self->_sendRequest('RelistItem', \@request, \%returnFields);

    return $result;
}

#========================================================================================
# §function     _getEbayItemFromAuction
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _getEbayItemFromAuction( $Auction );
#----------------------------------------------------------------------------------------
# §description  collects SOAP Data structure to upload an auction.
#----------------------------------------------------------------------------------------
# §input        $Auction | auction | object
# §input        $Relisting | true if run by relisting call | boolean (optinal)
# §return       $aItem | SOAP AddItem datastructure | array.ref
#========================================================================================
sub _getEbayItemFromAuction { ##no critic qw(ProhibitExcessComplexity)
    my $self = shift;
    my ($Auction, $Relisting) = @_;

    my $ShopEnv = $Auction->get('EbayShopEnvironment');
    my $Env = $ShopEnv->get('EbayEnvironment');

    my $hAuctionInfo = $Auction->get([qw(
        Name            SubTitle               Description
        Quantity        StartPrice             ListingType
        HitCounter      EbayCategoryId         DispatchTimeMax
        IsPrivate       Duration               SecondCategoryId
        EbayCategoryId  EbayUUID               IsBuyItNowEnabled
        BuyItNowPrice   AcceptsReturnPolicy    AcceptsReturnPolicyText
        IsGetItFast     EbayStoreCategoryId    EbayStoreCategory2Id
        IsBoldTitle     IsHighlight            IsBordered
        IsFeatured      IsHomePageFeatured     PictureUrl
        Product         AuctionPaymentMethods
        Gallery         AuctionShippingMethods AuctionShippingLocations
        ConditionID     IsReviseInventory      IsBestOfferEnabled
        BestOfferAutoAcceptPrice    MinimumBestOfferPrice ReturnsWithin
        BusinessPolicyShippingID BusinessPolicyPaymentID BusinessPolicyReturn_PolicyID
        EbayCatalogAddShopProductDescription EbayCatalogUseStockImages
        EbayCatalogBrandMappingRequired EbayCatalogMappingRequired
        )]);

    my $hShopEnvInfo = $ShopEnv->get([qw(
        EbayCountry     EbayRegion
        EbayLocation    EbayPostalCode
        PaypalAccount
    )]);

    my $currencyId = $Env->get('CurrencyID');

    my $description = $hAuctionInfo->{'Description'};

     #  EPG-22522 Replace Baseprice tag.
    my $processedContent = '';
    my $UserTemplate = $Auction->get('EbayUserTemplate');
    my $Shop = $Auction->getSite();
    if ((defined $UserTemplate || $Shop->get('IsUnityRegistered')) && $Auction->get('ListingType') eq 'FixedPriceItem'){
        my $EbayEnvironment = $Auction->get('EbayShopEnvironment')->get('EbayEnvironment');
        my $Product = $Auction->get('Product');
        my $Price = $Auction->get('StartPrice');
        $processedContent = _ReplaceBasePrice($Product, $EbayEnvironment, $Price, $Auction->get('Description'));

    } else {
        $processedContent = _DeleteBasePrice($description);
    }
    $description = $processedContent;

    my $Product = $hAuctionInfo->{'Product'};
    my $ListingType = $hAuctionInfo->{'ListingType'};
    my $aItem = [
        SOAP::Data->name('Site'                  )->value($Env->get('EbaySiteCode')),
        SOAP::Data->name('ListingType'           )->value($ListingType),
        SOAP::Data->name('Title'                 )->value(QuoteXML($hAuctionInfo->{'Name'})),
        SOAP::Data->name('SubTitle'              )->value(QuoteXML($hAuctionInfo->{'SubTitle'})),
        SOAP::Data->name('Currency'              )->value($currencyId),
        SOAP::Data->name('Country'               )->value($hShopEnvInfo->{'EbayCountry'}),
        SOAP::Data->name('RegionID'              )->value($hShopEnvInfo->{'EbayRegion'}),
        SOAP::Data->name('Location'              )->value(QuoteXML($hShopEnvInfo->{'EbayLocation'})),
        SOAP::Data->name('PrimaryCategory'       )->value({'CategoryID' => $hAuctionInfo->{'EbayCategoryId'}}),
        SOAP::Data->name('CategoryMappingAllowed')->value('true'),
        SOAP::Data->name('HitCounter'            )->value($hAuctionInfo->{'HitCounter'}),
        SOAP::Data->name('PrivateListing'        )->value(qw (false true)[$hAuctionInfo->{'IsPrivate'}]),
        SOAP::Data->name('DispatchTimeMax'       )->value($hAuctionInfo->{'DispatchTimeMax'}),
    ];

    if(!$Product->get('IsMaster')){
        if ($Shop->get('EbayStockDecreaseType') eq 'OnAuctionEnd' && $ListingType eq 'FixedPriceItem' && $hAuctionInfo->{'IsReviseInventory'}) {
            my $StockLevel = $Product->get('StockLevel');
            $hAuctionInfo->{'Quantity'} = 0;
            if ($StockLevel > 0) {
                $hAuctionInfo->{'Quantity'} = $StockLevel
            }
            $Auction->set({'Quantity' => $hAuctionInfo->{'Quantity'}});

            push @$aItem, SOAP::Data->name('OutOfStockControl')->value('true');
        }
        push @$aItem, SOAP::Data->name('Quantity'              )->value($hAuctionInfo->{'Quantity'});
        push @$aItem, SOAP::Data->name('StartPrice'            )->value($hAuctionInfo->{'StartPrice'})->attr({'currencyID' => $currencyId});
    }

    # add the description. Can be enabled on the assignment page
    if($hAuctionInfo->{'EbayCatalogBrandMappingRequired'} or $hAuctionInfo->{'EbayCatalogMappingRequired'}){
        push @$aItem, SOAP::Data->name('Description')->value(QuoteXML($description)) if $hAuctionInfo->{'EbayCatalogAddShopProductDescription'};
    } else {
        push @$aItem, SOAP::Data->name('Description')->value(QuoteXML($description));
    }

    if (defined $hShopEnvInfo->{'EbayPostalCode'}) {
        push @$aItem, SOAP::Data->name('PostalCode')->value(QuoteXML($hShopEnvInfo->{'EbayPostalCode'}));
    }

    my @AdditionalInfosForItemSpecifics;
    # EAN / UPC / Catalog information
    if ($Product) {
        my @ProductListingDetails;
        $self->_getProductListingDetails($Auction, $Product, \@ProductListingDetails, \@AdditionalInfosForItemSpecifics);

        if (scalar @ProductListingDetails) {
            push @$aItem, SOAP::Data->name('ProductListingDetails')->value(\SOAP::Data->value(@ProductListingDetails));
        }
    }

    if ($hAuctionInfo->{'Duration'} == DURATION_GOOD_TILL_CANCELED){
        push @$aItem, SOAP::Data->name('ListingDuration')->value('GTC');
    } else {
        push @$aItem, SOAP::Data->name('ListingDuration')->value('Days_'.$hAuctionInfo->{'Duration'});
    }

    if (defined $hAuctionInfo->{'SecondCategoryId'}) {
        push @$aItem, SOAP::Data->name('SecondaryCategory')->value({'CategoryID' => $hAuctionInfo->{'SecondCategoryId'}});
    }

    if (!$Relisting && defined $hAuctionInfo->{'EbayUUID'}) {
        push @$aItem, SOAP::Data->name('UUID')->value($hAuctionInfo->{'EbayUUID'});
    }
    else {
        push @$aItem, SOAP::Data->name('ItemID')->value($Auction->get('ExternalAuctionId'));
    }

    if ($hAuctionInfo->{'IsBuyItNowEnabled'}) {
        push @$aItem, SOAP::Data->name('BuyItNowPrice')->attr({'currencyID' => $currencyId})->value($hAuctionInfo->{'BuyItNowPrice'});
    }

    if ($hAuctionInfo->{'ConditionID'}) {
        push @$aItem, SOAP::Data->name('ConditionID')->value($hAuctionInfo->{'ConditionID'});
    }

    # add item specifics here
    my $aItemSpecifics = $Auction->getItemSpecifics();
    if (ref $aItemSpecifics eq 'ARRAY') {
        my $hReorganizedSpecifics = _reorganizeItemSpecifics($aItemSpecifics, \@AdditionalInfosForItemSpecifics);

        my @NameValueList;
        foreach my $ItemSpecificName (keys %$hReorganizedSpecifics){
            my $Name = SOAP::Data->name('Name')->value(QuoteXML($ItemSpecificName));
            my @Values = map { SOAP::Data->name('Value')->value(QuoteXML($_)) } @{$hReorganizedSpecifics->{$ItemSpecificName}};
            push @NameValueList, SOAP::Data->name('NameValueList')->value(\SOAP::Data->value(($Name), @Values));
        }
        push @$aItem, SOAP::Data->name('ItemSpecifics')->value(\SOAP::Data->value(@NameValueList)) if (scalar @NameValueList);
    }


    if ($ShopEnv->isBusinessPolicyActivated($ShopEnv->getSellerProfiles)){
        my $hSellerProfileData = $self->_getSellerProfileData($hAuctionInfo);

        push @$aItem, $self->_createBusinessPolicySoap($hSellerProfileData);
    }
    else {
        # add return policy
        my $ReturnAccepted = $hAuctionInfo->{'AcceptsReturnPolicy'} ? 'ReturnsAccepted' : 'ReturnsNotAccepted'  ;
        my $AcceptsReturnPolicyText = $hAuctionInfo->{'AcceptsReturnPolicyText'};

        my @container;
        push @container,SOAP::Data->name('ReturnsAcceptedOption')->value($ReturnAccepted);
        push @container,SOAP::Data->name('ReturnsWithinOption')->value($hAuctionInfo->{'ReturnsWithin'}) if defined $hAuctionInfo->{'ReturnsWithin'};
        push @container,SOAP::Data->name('Description')->value(QuoteXML($AcceptsReturnPolicyText)) if $hAuctionInfo->{'AcceptsReturnPolicy'};

        # always add return policy (EPG-12042)
        push @$aItem, SOAP::Data->name('ReturnPolicy')->value(\SOAP::Data->value(@container));

    }

    # add getItFast
    if ($Env->get('HasGetItFast') and $hAuctionInfo->{'IsGetItFast'}) {
        push @$aItem, SOAP::Data->name('GetItFast')->value('true');
    } else {
        push @$aItem, SOAP::Data->name('GetItFast')->value('false');
    }

    # add custom store categories
    my $StoreCategoryID = $hAuctionInfo->{'EbayStoreCategoryId'};
    my $StoreCategory2ID = $hAuctionInfo->{'EbayStoreCategory2Id'};

    if (defined $StoreCategoryID or defined $StoreCategory2ID) {
        push @$aItem, SOAP::Data->name('Storefront')->value({
            'StoreCategoryID'  => ( not defined $StoreCategoryID  or $StoreCategoryID  eq 'dummy' or $StoreCategoryID   == 0 ) ? undef : $StoreCategoryID,
            'StoreCategory2ID' => ( not defined $StoreCategory2ID or $StoreCategory2ID eq 'dummy' or $StoreCategory2ID  == 0 ) ? undef : $StoreCategory2ID,
        });
    }

    # add best offer options if we have FixPriceItems
    if ($Auction->get('ListingType') eq 'FixedPriceItem') {

        my $IsBestOfferEnabled = $hAuctionInfo->{'IsBestOfferEnabled'};
        # best offer is enabled, add fields....
        if ($IsBestOfferEnabled) {
            # enable best offer
            push @$aItem, SOAP::Data->name('BestOfferDetails')->value({
                'BestOfferEnabled' =>  SOAP::Data->value('true')
            });

            #add prices
            my $BestOfferAutoAcceptPrice = $hAuctionInfo->{'BestOfferAutoAcceptPrice'};
            my $MinimumBestOfferPrice = $hAuctionInfo->{'MinimumBestOfferPrice'};

            push @$aItem, SOAP::Data->name('ListingDetails')->value({
                'BestOfferAutoAcceptPrice'  => defined $BestOfferAutoAcceptPrice  ? SOAP::Data->value( $BestOfferAutoAcceptPrice )->attr({'currencyID' => $currencyId}) : undef ,
                'MinimumBestOfferPrice'     => defined $MinimumBestOfferPrice  ? SOAP::Data->value( $MinimumBestOfferPrice )->attr({'currencyID' => $currencyId}) : undef ,
            });
        }
    }

    # add VATDetails (inkl. Mwst., VAT)
    if (! $Shop->get('VATreleased')) {
        my $AuctionProduct = $hAuctionInfo->{'Product'};
        if (defined $AuctionProduct) {
            my $TaxMatrix = $Shop->get('TaxMatrix');
            my $TaxClass = $AuctionProduct->getWithParent('TaxClass');
            if (defined $TaxClass) {
                my $DefaultTaxRate = $TaxMatrix->taxRate($TaxMatrix->defaultTaxArea, $TaxClass);
                push @$aItem, SOAP::Data->name('VATDetails')->value({
                    'VATPercent'     => $DefaultTaxRate * 100, ## no critic qw(ProhibitMagicNumbers)
                });
            }
        }
    }

    if ($hShopEnvInfo->{'PaypalAccount'}) {
        push @$aItem, SOAP::Data->name('PayPalEmailAddress')->value($hShopEnvInfo->{'PaypalAccount'});
    }
    if ($hAuctionInfo->{'IsBoldTitle'}) {
        push @$aItem, SOAP::Data->name('ListingEnhancement')->value('BoldTitle'),
    }
    if ($hAuctionInfo->{'IsHighlight'}) {
        push @$aItem, SOAP::Data->name('ListingEnhancement')->value('Highlight'),
    }
    if ($hAuctionInfo->{'IsBordered'}) {
        push @$aItem, SOAP::Data->name('ListingEnhancement')->value('Border'),
    }
    if ($hAuctionInfo->{'IsFeatured'}) {
        push @$aItem, SOAP::Data->name('ListingEnhancement')->value('Featured'),
    }
    if ($hAuctionInfo->{'IsHomePageFeatured'}) {
        push @$aItem, SOAP::Data->name('ListingEnhancement')->value('HomePageFeatured'),
    }

    my $aVendorHostedPicture = [];

    #use catalog images if needed
    unless($hAuctionInfo->{'EbayCatalogUseStockImages'}){
        if (defined $hAuctionInfo->{'PictureUrl'} and $hAuctionInfo->{'PictureUrl'} =~ /\S/) {
            #since api v.393 the first of SelfHostedURL will be used as PictureURL
            push @$aVendorHostedPicture, (
                SOAP::Data->name('PictureURL' => $hAuctionInfo->{'PictureUrl'}),
                SOAP::Data->name('PhotoDisplay' => 'None'),
            );
        }
    }

    if (defined $hAuctionInfo->{'Gallery'} and $hAuctionInfo->{'Gallery'} =~ /\S/) {
        push @$aVendorHostedPicture, (
            SOAP::Data->name('GalleryType' => $hAuctionInfo->{'Gallery'}),
            #GalleryDuration  can be also added since 581
        );
    }
    if (scalar @$aVendorHostedPicture) {
        push @$aItem, SOAP::Data->name('PictureDetails')->value(\SOAP::Data->value(@$aVendorHostedPicture));
    }

    my $CODCost;
    # payment
    foreach my $payment (@{$hAuctionInfo->{'AuctionPaymentMethods'}}) {
        $CODCost = $CODCost//$payment->get('CODCost');
        push @$aItem, SOAP::Data->name('PaymentMethods'        )->value( $payment->get('EbayPaymentMethod')->alias);
    }

    # shipping
    my $prio = 1;
    my @shippings = ();

    if (defined $CODCost) {
        push @shippings, SOAP::Data->name('CODCost')->value($CODCost)->attr({'currencyID' => $currencyId});
    }

    foreach my $ship (@{$hAuctionInfo->{'AuctionShippingMethods'}}) {
        my $hInfoShipping = {
            'ShippingService'               => $ship->get('EbayShippingMethod')->alias,
            'ShippingServiceCost'           => SOAP::Data->value( $ship->get('Price') )->attr({'currencyID' => $currencyId}),
            'ShippingServicePriority'       => $prio++,
        };
        my $AdditionalPrice = $ship->get('AdditionalPrice');
        $hInfoShipping->{'ShippingServiceAdditionalCost'} =
            SOAP::Data->value( $AdditionalPrice )->attr({'currencyID' => $currencyId})
            if (defined $AdditionalPrice);
        push @shippings, SOAP::Data->name('ShippingServiceOptions')->value($hInfoShipping);

    }

    # shipping location
    my $priority = 1;
    foreach my $ShippingLocation (@{$hAuctionInfo->{'AuctionShippingLocations'}}) {
        my $hInfoShipping = {
            'ShippingService'               => $ShippingLocation->get('EbayShippingMethod')->alias,
            'ShippingServiceCost'           => SOAP::Data->value( $ShippingLocation->get('Price') )->attr({'currencyID' => $currencyId}),
            'ShippingServicePriority'       => $priority++,
            'ShipToLocation'                => $ShippingLocation->get('EbayShippingLocation')->alias,
        };
        my $AdditionalPrice = $ShippingLocation->get('AdditionalPrice');
        $hInfoShipping->{'ShippingServiceAdditionalCost'} =
            SOAP::Data->value( $AdditionalPrice  )->attr({'currencyID' => $currencyId})
            if (defined $AdditionalPrice);
        push @shippings, SOAP::Data->name('InternationalShippingServiceOption')->value($hInfoShipping);
    }
    push @$aItem, SOAP::Data->name('ShippingDetails')->value(\SOAP::Data->value(@shippings));

    # Multi Variations Auction
    if($Product->get('IsMaster')){
        if ($Shop->get('EbayStockDecreaseType') eq 'OnAuctionEnd' && $ListingType eq 'FixedPriceItem' && $hAuctionInfo->{'IsReviseInventory'}) {
            push @$aItem, SOAP::Data->name('OutOfStockControl')->value('true');
        }

        my $haVariationSoap = $self->_getVariationSOAP($Auction, $currencyId);
        push @$aItem, SOAP::Data->name('Variations')->value(\SOAP::Data->value(@{$haVariationSoap}));
    }

    return $aItem;
}


sub _getProductListingDetails {
    my $self = shift;
    my ($Auction,$Product,$aProductListingDetails, $aAdditionalInfosForItemSpecifics) = @_;

    my $hProductInfo = $Product->getWithParent([qw( Manufacturer UPCEAN Manufacturer ManufacturerSKU IsUPCEANValid )]);

    if (! $Product->get('IsMaster')) {

        if (isValid($hProductInfo->{'UPCEAN'})) {

            # delete all dashes and spaces
            $hProductInfo->{'UPCEAN'} =~ s/-//ig;
            $hProductInfo->{'UPCEAN'} =~ s/ //ig;

            my $isbn10 = CheckDigits('isbn');
            my $isbn13 = CheckDigits('isbn13');
            my $ean    = CheckDigits('ean');
            my $upc    = CheckDigits('upc');

            if ($isbn10->is_valid($hProductInfo->{'UPCEAN'}) or $isbn13->is_valid($hProductInfo->{'UPCEAN'})) {
                push @$aProductListingDetails, SOAP::Data->name('ISBN')->value(QuoteXML($hProductInfo->{'UPCEAN'}));
                push @$aAdditionalInfosForItemSpecifics, { 'Name' => 'ISBN', 'Value' => $hProductInfo->{'UPCEAN'}};
            }
            if ($upc->is_valid($hProductInfo->{'UPCEAN'})) {
                push @$aProductListingDetails, SOAP::Data->name('UPC')->value(QuoteXML($hProductInfo->{'UPCEAN'}));
                push @$aAdditionalInfosForItemSpecifics, { 'Name' => 'UPC', 'Value' => $hProductInfo->{'UPCEAN'}};
            }
            if ($ean->is_valid($hProductInfo->{'UPCEAN'})) {
                push @$aProductListingDetails, SOAP::Data->name('EAN')->value(QuoteXML($hProductInfo->{'UPCEAN'}));
                push @$aAdditionalInfosForItemSpecifics, { 'Name' => 'EAN', 'Value' => $hProductInfo->{'UPCEAN'}};
            }
        }
        # if the product has no EAN/ISBN/UPC we need to send eBay an information about this.
        # see AD-2267
        else {
            my $siteId = $self->environment->get('EbaySiteId');
            my $defaultEANValues = EBAY_DEFAULT_EAN_VALUE();
            my $defaultISBNValues = EBAY_DEFAULT_ISBN_VALUE();
            my $defaultUPCValues = EBAY_DEFAULT_UPC_VALUE();

            if (exists $defaultEANValues->{$siteId}) {
                push @$aProductListingDetails, SOAP::Data->name('EAN')->value($defaultEANValues->{$siteId});
            }
            if (exists $defaultISBNValues->{$siteId}) {
                push @$aProductListingDetails, SOAP::Data->name('ISBN')->value($defaultISBNValues->{$siteId});
            }
            if (exists $defaultUPCValues->{$siteId}) {
                push @$aProductListingDetails, SOAP::Data->name('UPC')->value($defaultUPCValues->{$siteId});
            }
        }

        # we have to send also the ePid for th catalog product if exists as "ProductReferenceID"
        my $EbayCatalogProductID = $Auction->get('EbayCatalogProductID');
        push @$aProductListingDetails, SOAP::Data->name('ProductReferenceID')->value($EbayCatalogProductID)
             if defined $EbayCatalogProductID;

    }

    # manufacturer / sku
    if (isValid($hProductInfo->{'Manufacturer'} && isValid($hProductInfo->{'ManufacturerSKU'}))) {
        my @BrandMPN;
        push @BrandMPN, SOAP::Data->name('Brand')->value(QuoteXML($hProductInfo->{'Manufacturer'}));
        push @BrandMPN, SOAP::Data->name('MPN')->value(QuoteXML($hProductInfo->{'ManufacturerSKU'}));
        push @$aProductListingDetails, SOAP::Data->name('BrandMPN')->value(\SOAP::Data->value(@BrandMPN));

        push @$aAdditionalInfosForItemSpecifics, { 'Name' => 'Brand', 'Value' => $hProductInfo->{'Manufacturer'}};
        push @$aAdditionalInfosForItemSpecifics, { 'Name' => 'MPN', 'Value' => $hProductInfo->{'ManufacturerSKU'}};
    }
}

sub _reorganizeItemSpecifics {
    my ($aEbaySpecifics, $aProductSpecifics) = @_;

    # take always Brand from EbayWizard, if undefined take Product Manufacturer(Brand), see EPC-41
    my $Brand = 'Brand';
    my @aBrandEbayWizard = grep { $_->{Name} eq $Brand} @$aEbaySpecifics;
    my @aProductSpecifics = @aBrandEbayWizard ? grep { $_->{Name} ne $Brand} @$aProductSpecifics : @$aProductSpecifics;
    my @aData = (@$aEbaySpecifics, @aProductSpecifics);

    my %Target;
    for my $Element ( @aData ) {
        my $k = $Element->{Name};
        my $v = $Element->{Value};
        my $tmp = $Target{ $k } //= [];
        push @$tmp, $v;
    }
    return \%Target;
}

#========================================================================================
# §function     getEbayDetails
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getEbayDetails($DetailName);
#----------------------------------------------------------------------------------------
# §description  download countries, regions, shippings.
# §input        $DetailName | name of site detail | string
# §return       $result | result of _sendRequest | ref.hash
#========================================================================================
sub getEbayDetails {
    my $self = shift;
    my ($DetailName) = @_;
    Error('FalseDetail') unless ElementOf($DetailName, EBAY_DETAILS);
    my @request = (
        SOAP::Data->name('Site')->value($self->environment->get('EbaySiteCode')),
        SOAP::Data->name('DetailName')->value($DetailName),
    );

    my %returnFields = ('Fields' => [],
                        'Arrays' => [$DetailName],
    );

    my $result = $self->_sendRequest('GeteBayDetails', \@request, \%returnFields);

    return $result;
}

#========================================================================================
# §function     getEbayOfficialTime
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getEbayOfficialTime($NoToken);
#----------------------------------------------------------------------------------------
# §input        $NoToken | connect without using the token | int
# §return       $result | with following keys:
#               <ul>
#                   <li>Timestamp - ebay timestamp - string
#                   <li>Ack - "Success" if successful - string
#                   <li>CorrelationID - ebay correlation id - string
#                   <li>Version - ebay api version - string
#                   <li>Build - ebay build no - string
#               </ul> | ref.hash
#----------------------------------------------------------------------------------------
# §description  get the ebay official time
#========================================================================================
sub getEbayOfficialTime {
    my $self = shift;
    my ($NoToken) = @_;

    my @request = ();
    my %returnFields = ();

    my $result = $self->_sendRequest('GeteBayOfficialTime', \@request, \%returnFields,$NoToken);

    return $result;
}

#========================================================================================
# §function     getUser
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getUser($aParams);
# §example      my $bIsStoreOwner = $RemoteEbay->getUser(['StoreOwner'])->{'StoreOwner'};
#----------------------------------------------------------------------------------------
# §description  Get informations about the eBay user
#               (http://developer.ebay.com/DevZone/XML/docs/Reference/eBay/index.html)
#----------------------------------------------------------------------------------------
# §input        $aParams | array with fields to get | ref.array
# §return       $hresult | hash with results | ref.hash
#========================================================================================
sub getUser {
    my $self = shift;
    my ($aParams) = @_;

    return {} unless ref $aParams eq 'ARRAY';

    my %returnFields = ('Fields' => $aParams,
                        'Arrays' => [],
    );
    my $hresult = $self->_sendRequest('GetUser', [], \%returnFields);

    return $hresult;
}

#========================================================================================
# §function     getOrders
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getOrders($aParams);
# §example      my $bIsStoreOwner = $RemoteEbay->getOrders(['StoreOwner'])->{'StoreOwner'};
#----------------------------------------------------------------------------------------
# §description  Get list of ebay orders (transactions with combined payment)
#----------------------------------------------------------------------------------------
# §input        $aParams | array with fields to get | ref.array
# §return       $hresult | hash with results | ref.hash
#========================================================================================
sub getOrders {
    my $self = shift;
    my ($hParams) = @_;
    my %returnFields = ();

    my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();
    my $now = DateTime->now;

    my $hresult = $self->_sendRequest('GetOrders', [
        SOAP::Data->name('OrderStatus')->value($hParams->{'OrderStatus'}//'Active'),
        SOAP::Data->name('CreateTimeFrom')->value($parser->format_datetime($hParams->{'CreateTimeFrom'})),
        SOAP::Data->name('CreateTimeTo')->value($parser->format_datetime($now)),
        SOAP::Data->name('OrderRole')->value('Seller')
    ], \%returnFields);
    return $hresult;
}

#========================================================================================
# §function     completeSale
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->completeSale($externalTransactionID, $ItemID, $hParams);
# §example      $RemoteEbay->completeSale($t, $i, {'Paid'=> 1}};
#----------------------------------------------------------------------------------------
# §description  Complete a transaction on ebay side
#----------------------------------------------------------------------------------------
# §input        $externalTransactionID | ebay transaction id | integer
# §input        $ItemID | ebay item id | integer
# §input        $hParams | options. used keys: <ul>
#                 <li>Paid - paid status - boolean</li>
#               </ul> | ref.hash
#========================================================================================
sub completeSale {
    my $self = shift;
    my ($externalTransactionID, $ItemID, $hParams) = @_;
    my %returnFields = ();

    my @request = (
        SOAP::Data->name('ItemID')->value($ItemID),
        SOAP::Data->name('TransactionID')->value($externalTransactionID),
        SOAP::Data->name('Paid')->value($hParams->{'Paid'})
    );

    $self->_sendRequest('CompleteSale', \@request, \%returnFields);
}

#========================================================================================
# §function     reviseInventoryStatus
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->reviseInventoryStatus($externalAuctionId, $SKU, $quantity);
# §example      $RemoteEbay->reviseInventoryStatus($externalAuctionId, undef, $quantity};
#----------------------------------------------------------------------------------------
# §description  Revise a transaction's quantity for a automatic stocklevel update
#               Variation also are identified by their SKU/ProductNo
#----------------------------------------------------------------------------------------
# §input        $externalAuctionId | ebay transaction id | integer
# §input        $SKU | product alias as variations identifier | string
# §input        $quantity | ebay item id | integer
#========================================================================================
sub reviseInventoryStatus {
    my $self = shift;
    my ($externalAuctionId, $SKU, $quantity) = @_;
    my %returnFields = ();

    my @Item = (
        SOAP::Data->name('ItemID')->value($externalAuctionId),
        SOAP::Data->name('Quantity')->value($quantity)
    );

    if ($SKU) {
        push @Item, SOAP::Data->name('SKU')->value($SKU);
    }

    my @request = (
        SOAP::Data->name('InventoryStatus')->value(\@Item)
    );

    $self->_sendRequest('ReviseInventoryStatus', \@request, \%returnFields);
}

#========================================================================================
# §function     setNotificationPreferences
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->setNotificationPreferences( $aEnableEvents, $aDisableEvents, $URL );
#----------------------------------------------------------------------------------------
# §description  set notification preferences
#----------------------------------------------------------------------------------------
# §input        $aEnableEvents  | list of events to enable    | string.array.ref
# §input        $aDisableEvents | list of events to disable   | string.array.ref
# §input        $URL            | url to send notification to | string
#========================================================================================
sub setNotificationPreferences {
    my $self = shift;
    my ($aEnableEvents, $aDisableEvents, $URL) = @_;
    my %returnFields;
    my @request;

    push @request, SOAP::Data->name('ApplicationDeliveryPreferences')->value(
    \SOAP::Data->value(
        SOAP::Data->name('ApplicationEnable')->value('Enable'),
        SOAP::Data->name('DeliveryURLDetails')->value(
        \SOAP::Data->value(
            SOAP::Data->name('DeliveryURL')->value(QuoteXML($URL)),
            SOAP::Data->name('Status')->value('Enable')
        ))
    ));

    my @events;

    foreach my $Event (@$aEnableEvents){
        push @events, SOAP::Data->name('NotificationEnable')->value(
            \SOAP::Data->value(
                SOAP::Data->name('EventEnable')->value('Enable'),
                SOAP::Data->name('EventType')->value($Event),
            ));
    }

    foreach my $Event (@$aDisableEvents){
        push @events,
            SOAP::Data->name('NotificationEnable')->value(
            \SOAP::Data->value(
                SOAP::Data->name('EventEnable')->value('Disable'),
                SOAP::Data->name('EventType')->value($Event),
            ));
    }

    push (@request, SOAP::Data->name('UserDeliveryPreferenceArray')->value(
    \SOAP::Data->value(@events))) if (scalar @events);

    push (@request, SOAP::Data->name('DeliveryURLName')->value($self->shopEnvironment->get('EbayToken')));

    $self->_sendRequest('SetNotificationPreferences', \@request, \%returnFields);
}

#========================================================================================
# §function     getNotificationPreferences
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getNotificationPreferences( $PreferenceLevel );
#----------------------------------------------------------------------------------------
# §description  Retrieves notification preferences
#----------------------------------------------------------------------------------------
# §input        $PreferenceLevel | preference level (User/Application etc.) | string
# §return       $hResult | soap response | hash.ref
#========================================================================================
sub getNotificationPreferences {
    my $self = shift;
    my ($PreferenceLevel) = @_;
    my %returnFields;

    my $hresult = $self->_sendRequest('SetNotificationPreferences',
    [
        SOAP::Data->name('PreferenceLevel')->value($PreferenceLevel//'User')
    ], \%returnFields);

    LogDebug( 'NotificationPreferences', $hresult);
    return $hresult;
}

#========================================================================================
# §function     getCategoryFeatures
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getCategoryFeatures( $CategoryID, $aFeatureIDs, $ViewAllNodes, $AllFeaturesForCategory );
#----------------------------------------------------------------------------------------
# §description  calls GetCategoryFeatures. All parameters are optional
#----------------------------------------------------------------------------------------
# §input        $CategoryID,  | ebay category id | int
# §input        $aFeatureIDs    | ebay feature ids | ref.array.string
# §input        $ViewAllNodes | true to receive site defaults  | boolean
# §input        $AllFeaturesForCategory | true to receive all values for category, even
#               the default values | boolean
# §input        $hReturnFields | limit returned fields to a sane list of values. Default: <ul>
#                 <li>Fields => [ ItemSpecificsEnable ConditionValues ] </li>
#                 <li>Arrays => [ Condition ] </li>
#               </ul> | ref.hash
# §return       $GetCategoryFeaturesResponse | soap response | hash.ref
#========================================================================================
sub getCategoryFeatures {
    my $self = shift;
    my ($CategoryID, $aFeatureIDs, $ViewAllNodes, $AllFeaturesForCategory, $hReturnFields) = @_;
    $hReturnFields //= {'Fields' => [qw ( ItemSpecificsEnabled ConditionValues ProductRequiredEnabled ) ],
                        'Arrays' => [qw( Condition )]
    };
    my @request;

    push (@request, SOAP::Data->name('CategoryID',$CategoryID)) if (defined $CategoryID);
    if (ref $aFeatureIDs eq 'ARRAY') {
        push (@request, SOAP::Data->name('FeatureID', $_)) foreach (@$aFeatureIDs);
    }
    push (@request, SOAP::Data->name('ViewAllNodes','true')) if ($ViewAllNodes);
    push (@request, SOAP::Data->name('AllFeaturesForCategory','true')) if ($AllFeaturesForCategory);
    push (@request, SOAP::Data->name('DetailLevel','ReturnAll'));

    my $hResult = $self->_sendRequest('GetCategoryFeatures',
        \@request, $hReturnFields);

    return $hResult;
}

#========================================================================================
# §function     getCategorySpecifics
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getCategorySpecifics( $CategoryID );
#----------------------------------------------------------------------------------------
# §description  Retrieve custom item specifics for a category.
#----------------------------------------------------------------------------------------
# §input        $CategoryID | category id | integer
# §return       $hResult | soap response | hash.ref
#========================================================================================
sub getCategorySpecifics {
     my $self = shift;
     my ($CategoryID) = @_;

    my %returnFields = ('Fields' => [qw ( Recommendations ) ],
                        'Arrays' => [qw( NameRecommendation ValueRecommendation )]);

    my @request;
    push (@request, SOAP::Data->name('CategoryID',$CategoryID));
    push (@request, SOAP::Data->name('MaxValuesPerName','500'));

    my $hResult = $self->_sendRequest('GetCategorySpecifics',
        \@request, \%returnFields);

    foreach my $hNameRecomm (@{$hResult->{'NameRecommendation'}}){
        next unless exists $hNameRecomm->{'ValueRecommendation'};
        $hNameRecomm->{'ValueRecommendation'} = [$hNameRecomm->{'ValueRecommendation'}] unless ref $hNameRecomm->{'ValueRecommendation'} eq 'ARRAY';
    }

    return $hResult;
}

#========================================================================================
# §function     getUserPreferences
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getUserPreferences( $aShowPreferenceCategory );
# §example      my $hPrefs = $self->getUserPreferences( ['ShowSellerPaymentPreferences'] );
#----------------------------------------------------------------------------------------
# §description  Retrieve user preferences
#----------------------------------------------------------------------------------------
# §input        $aShowPreferenceCategory | list of preference categories | string.array.ref
# §return       $hResult | soap response | hash.ref
#========================================================================================
sub getUserPreferences {
    my $self = shift;
    my ($aShowPreferenceCategory) = @_;

    my %returnFields = ('Fields' => [qw (
        BidderNoticePreferences           CombinedPaymentPreferences
        CrossPromotionPreferences         EmailShipmentTrackingNumberPreference
        EndOfAuctionEmailPreferences      ProStoresPreference
        PurchaseReminderEmailPreferences  SellerExcludeShipToLocationPreferences
        SellerFavoriteItemPreferences     SellerPaymentPreferences
        UnpaidItemAssistancePreferences   SellerProfilePreferences
        ) ]);

    my @request;
    foreach my $Category (@{$aShowPreferenceCategory}){
        push (@request, SOAP::Data->name($Category,'true'));
    }

    my $hResult = $self->_sendRequest('GetUserPreferences',
        \@request, \%returnFields);

    return $hResult;
}

#========================================================================================
# §function     setUserPreferencesDisplayPayNowButton
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->setUserPreferencesDisplayPayNowButton( $DisplayPayNowButtonSetting );
# §example      $self->setUserPreferencesDisplayPayNowButton( 'ShowPayNowButtonForAllPaymentMethods' );
#----------------------------------------------------------------------------------------
# §description  Set user preference: DisplayPayNowButton. This enables / disables ebays
#               platform integrated checkout functionality. For a list of valid options
#               see ebay API documentation for Type "DisplayPayNowButtonCodeType"
#----------------------------------------------------------------------------------------
# §input        $DisplayPayNowButtonSetting | pay now button setting | string
#========================================================================================
sub setUserPreferencesDisplayPayNowButton {
    my $self = shift;
    my ($DisplayPayNowButtonSetting) = @_;

    my @request;
    push (@request , SOAP::Data->name('SellerPaymentPreferences')->value(
        \SOAP::Data->value(
                SOAP::Data->name('DisplayPayNowButton')->value($DisplayPayNowButtonSetting)
    )));

    $self->_sendRequest('SetUserPreferences', \@request, {});
}


#========================================================================================
# §function     QuoteXML
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       my $quotedString = QuoteXML("xml string with characters like ? etc.");
#----------------------------------------------------------------------------------------
# §description  replace special characters with xml entities and return result string
#               workaround for encoding problems (EPG-7106)
#----------------------------------------------------------------------------------------
# §input        $str | string with special characters | string
# §return       $str | string encoded with xml entities | string
#========================================================================================
sub QuoteXML {
    my ($str) = @_;
    $str = DE_EPAGES::Core::API::String::QuoteXML($str);
    return unless defined $str;
    return join '', map{ ord($_)<128 ? $_ : sprintf('&#%d;', ord($_)) } split //, $str; ## no critic qw(ProhibitMagicNumbers)
}

#========================================================================================
# §function     _referencePrice
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       my $RefPrice  = _referencePrice($Product,$Shop, $Price);
#----------------------------------------------------------------------------------------
# §description  EPG-22522 Returns String with all data regarding reference price
#----------------------------------------------------------------------------------------
# §input        #Product | product object | object
# §input        #Shop    | Shop object | object
# §input        #Price   | price | integer
#----------------------------------------------------------------------------------------
# §return       undef
#========================================================================================
sub _referencePrice {
    my ( $Product, $EbayEnvironment, $Price ) = @_;
    my $RefPrice = $Product->getReferencePrice($EbayEnvironment->get('CurrencyID'), $Price);

    #ReferencePrice != #DisplayPrice same as in SF-Product.Content-Base-ImageUprightRight.html
    if ( defined $Price && defined $RefPrice ) {
        my $RefAmount = $Product->getWithParent('RefAmount');
        my $RefUnit   = $Product->getWithParent('RefUnit');
        my $Local     = $Product->getSite()->siteLocale($EbayEnvironment->get('LocaleID'));
        my $EnvironmentLanguageID = GetPKeyLanguageByCode(lc $EbayEnvironment->get('LanguageID'));

        if (defined $RefAmount && defined $RefUnit){
            return sprintf(
            '%d %s = %s',
            $RefAmount,
            $RefUnit->get('Abbreviation')
              // $RefUnit->get( 'NameOrAlias', $EnvironmentLanguageID ),
            $Local->formatMoney($RefPrice, $EbayEnvironment->get('CurrencyID'))
            );
        }
    }
    return undef;
}

sub _ReplaceBasePrice {
    my ($Product, $EbayEnvironment, $Price, $content) = @_;
    my $hvalues = {
            'Baseprice'    => _referencePrice( $Product, $EbayEnvironment, $Price),
        };

    my $processedContent = ReplaceTLEs($content , $hvalues );

    return $processedContent;
}

sub _DeleteBasePrice {
    my ($content) = @_;
    my $hvalues = {
            'Baseprice'    => undef,
        };

    my $processedContent = ReplaceTLEs($content , $hvalues );

    return $processedContent;
}

sub _getVariationSOAP{
    my $self = shift;
    my ($Auction, $currencyId ) = @_;
    my $LanguageID = GetPKeyLanguageByCode(lc $Auction->get('EbayShopEnvironment')->get('EbayEnvironment')->get('LanguageID'));

    # Add all variation and values to the variation temp for later use
    my @variationsTemp;
    my @variationSpecificsSet;

    # Go through the variation type values
    foreach my $variationTypeValues (@{$Auction->get('Product')->get('VariationTypeValues')}) {
        my @nameValueList = ();
        # Add the variation to temp array
        push @variationsTemp, $variationTypeValues->{'Attribute'};
        push @nameValueList, SOAP::Data->name('Name')->value($variationTypeValues->{'Attribute'}->get('NameOrAlias',$LanguageID));

        # Add all values to temp array
        foreach my $variationName (@{$variationTypeValues->{'Values'}}){
            my $variationValue = $variationName->get('ValueOrAlias',$LanguageID);
            push @nameValueList, SOAP::Data->name('Value')->value(QuoteXML($variationValue));
        }

        push @variationSpecificsSet, SOAP::Data->name('NameValueList')->value(\SOAP::Data->value(@nameValueList));
    }

    my @variations;
    push @variations, SOAP::Data->name('VariationSpecificsSet')->value(\SOAP::Data->value(@variationSpecificsSet));

    my $selectedVariationName = '';
    my @VariationPicturesTags;
    if(defined $Auction->get('PicVariationAttribute')){
        my $selectedVariationAttribute = $Auction->get('PicVariationAttribute');
        $selectedVariationName = $selectedVariationAttribute->get('NameOrAlias',$LanguageID);
        push @VariationPicturesTags, SOAP::Data->name('VariationSpecificName')->value($selectedVariationName);
    }

    my $Shop = $Auction->getSite();
    my $AllowWithoutStock = 0;
    if ($Shop->get('EbayStockDecreaseType') eq 'OnAuctionEnd' && $Auction->get('IsReviseInventory')) {
        $AllowWithoutStock = 1;
    }

    my %hAddedVariationValues;
    # Go through the subproducts and add the needed infos
    my $MasterQuantity = 0;
    foreach my $SubAuction ( @{$Auction->get('SubAuctions')} ) {
        my $Quantity = $SubAuction->get('Quantity') // 0;
        if ($Shop->get('EbayStockDecreaseType') eq 'OnAuctionEnd' && $Auction->get('IsReviseInventory')) {
            my $StockLevel = $SubAuction->get('Product')->get('StockLevel');
            $Quantity = 0;
            if ($StockLevel > 0) {
                $Quantity = $StockLevel
            }
            $SubAuction->set({'Quantity' => $Quantity});
            $MasterQuantity += $Quantity;
        }
        if ($Quantity > 0 || $AllowWithoutStock) {
            my @variationSpecifics = ();
            my $subProduct = $SubAuction->get('Product');
            foreach my $variationTemp (@variationsTemp){
                my $variationName = $variationTemp->get('NameOrAlias',$LanguageID);
                my $variationValue = $subProduct->get($variationTemp->alias)->get('ValueOrAlias',$LanguageID);

                my $hInfoNameValueList = {
                    'Name'            => QuoteXML($variationName),
                    'Value'           => QuoteXML($variationValue),
                };

                push @variationSpecifics, SOAP::Data->name('NameValueList')->value($hInfoNameValueList);

                # Add the first variation name of the first subproduct to the picture tag
                if(0 == scalar @VariationPicturesTags) {
                    $selectedVariationName = $variationName;
                    push @VariationPicturesTags, SOAP::Data->name('VariationSpecificName')->value($variationName);
                }

                # Add the url of the first item of each variation value
                if($selectedVariationName eq $variationName && !exists($hAddedVariationValues{$variationValue}) && defined $SubAuction->get('PictureUrl')) {
                    $hAddedVariationValues{$variationValue} = undef;
                    my @VariationSpecificPictureSetTags = ();
                    push @VariationSpecificPictureSetTags, SOAP::Data->name('VariationSpecificValue')->value(QuoteXML($variationValue));
                    push @VariationSpecificPictureSetTags, SOAP::Data->name('PictureURL')->value($SubAuction->get('PictureUrl'));
                    push @VariationPicturesTags , SOAP::Data->name('VariationSpecificPictureSet')->value(\SOAP::Data->value(@VariationSpecificPictureSetTags));
                }
            }

            my @ProductListingDetails;
            my @AdditionalInfosForItemSpecifics;
            $self->_getProductListingDetails($SubAuction, $subProduct, \@ProductListingDetails, \@AdditionalInfosForItemSpecifics);

            my $hInfoVariations = {
                'Quantity'      => $Quantity,
                'SKU'           => $SubAuction->get('Product')->get('Alias'),
                'StartPrice'    => SOAP::Data->value( $SubAuction->get('StartPrice') )->attr({'currencyID' => $currencyId}),
                'VariationSpecifics' => SOAP::Data->value(\SOAP::Data->value(@variationSpecifics)),
            };

            if (scalar @ProductListingDetails) {
                $hInfoVariations->{'VariationProductListingDetails'} = SOAP::Data->name('VariationProductListingDetails')->value(\SOAP::Data->value(@ProductListingDetails));
            }

            push @variations, SOAP::Data->name('Variation')->value($hInfoVariations);
        }
    }
    if ($Shop->get('EbayStockDecreaseType') eq 'OnAuctionEnd' && $Auction->get('IsReviseInventory')) {
        $Auction->set({'Quantity' => $MasterQuantity});
    }

    # Add Pictures
    push @variations , SOAP::Data->name('Pictures')->value(\SOAP::Data->value(@VariationPicturesTags));

    return \@variations;
}

#========================================================================================
# §function     _createBusinessPolicySoap
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       my $SellerProfiles    = $self->_createBusinessPolicySoap($hSellerProfileData)
#----------------------------------------------------------------------------------------
# §description  returns soap data of the seller profiles
#----------------------------------------------------------------------------------------
# §input        $hSellerProfileData | Hash with all infos of seller profile | hash.ref
# §return       $SellerProfiles | soap data for the seller profiles |  SOAP::DATA
#========================================================================================

sub _createBusinessPolicySoap {
    my $self = shift;
    my ($hSellerProfileData) = @_;

    my @SellerProfiles;

    my $SellerPaymentProfile    = $self->_getSellerProfile($hSellerProfileData,'Payment');
    my $ShippingPaymentProfile  = $self->_getSellerProfile($hSellerProfileData,'Shipping');
    my $ReturnPaymentProfile    = $self->_getSellerProfile($hSellerProfileData,'Return');

    push @SellerProfiles, $SellerPaymentProfile;
    push @SellerProfiles, $ShippingPaymentProfile;
    push @SellerProfiles, $ReturnPaymentProfile;

    return SOAP::Data->name('SellerProfiles')->value(\SOAP::Data->value(@SellerProfiles));
}

#========================================================================================
# §function     _getSellerProfile
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       my $SellerPaymentProfile    = $self->_getSellerProfile($hSellerProfileData,'Payment');
#----------------------------------------------------------------------------------------
# §description  returns soap data seller profile for a given profile type
#----------------------------------------------------------------------------------------
# §input        $hSellerProfileData | Hash with all infos of seller profile | hash.ref
# §input        $SellerProfileTypeName | Bussiness Policy Profile Type name in camel case | string
# §return       $SellerProfile | soap data for the specific seller profile type |  SOAP::DATA
#========================================================================================
sub _getSellerProfile {
    my $self = shift;
    my ($hSellerProfileData,$SellerProfileTypeName) = @_;

    my @container;

    my $SellerProfileValue = $SellerProfileTypeName eq 'Return' ? 'Return_Policy' : $SellerProfileTypeName;

    push @container,SOAP::Data->name($SellerProfileTypeName.'ProfileID')->value($hSellerProfileData->{uc($SellerProfileValue)}->{'ProfileID'});

    return SOAP::Data->name('Seller'.$SellerProfileTypeName.'Profile')->value(\SOAP::Data->value(@container));;
}

#========================================================================================
# §function     _getSellerProfileData
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       my $hSellerProfileData = $self->_getSellerProfileData($hAuctionInfo);
#----------------------------------------------------------------------------------------
# §description  returns a hash of the Seller Profile data for a given hash with Auction Infos
#----------------------------------------------------------------------------------------
# §input        $hAuctionInfo | Hash with all infos of the auction | hash.ref
# §return       $SellerProfile | Hash with all needed seller profile data | hash.ref
#========================================================================================
sub _getSellerProfileData {
    my $self = shift;
    my ($hAuctionInfo) = @_;

    return {
        'PAYMENT' => {
            'ProfileID'   => $hAuctionInfo->{'BusinessPolicyPaymentID'}
        },
        'SHIPPING' => {

            'ProfileID'   => $hAuctionInfo->{'BusinessPolicyShippingID'}
        },
        'RETURN_POLICY' => {

            'ProfileID'   => $hAuctionInfo->{'BusinessPolicyReturn_PolicyID'}
        }
    };
}

1;
