use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::Error qw( ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath LoadObject );
use DE_EPAGES::Core::API::PerlTools qw ( UsePackage );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName = 'Store';
    my $ShopAlias;
    my $Help;
    my $NoOutput;
    my $TestMode = 0;

    GetOptions(
        'help'          => \$Help,
        'nooutput'      => \$NoOutput,
        'passwd=s'      => \$Password,
        'storename=s'   => \$StoreName,
        'shop=s'        => \$ShopAlias,
        'test'        => \$TestMode,
    );

    usage() if $Help;

    unless ($StoreName) {
        print STDERR "Missing parameter storename\n\n";
        usage();
    }

    unless ($ShopAlias) {
        print STDERR "Missing parameter shop\n\n";
        usage();
    }

    my $time1 = time;

    RunOnStore(
        Store => $StoreName,
        DBPassword => $Password,
        Sub => sub  {

            # make sure that the cartridge is installed
            unless (ExistsCartridgeByPackage('DE_EPAGES::Ebay')) {
                print "Store: $StoreName : Cartridge Ebay is not installed.\n" unless ($NoOutput);
                return;
            }

            my $Shop;
            eval { $Shop = LoadObjectByPath('/Shops')->child($ShopAlias); };
            if ( ExistsError() ) {
                printf STDERR "[%s] WARN: Shop '%s' does not exist.\n", $StoreName, $ShopAlias;
                return;
            }
            my $aEbayProducts = _getProductWithRunningEbayAuctions($Shop);

            print "Running on Shop $ShopAlias \n Set Stocklevel:\n" unless ($NoOutput);

            # function to call, call hook function "OnAfterUpdateProduct" without TriggerHook
            my $FunctionName = "DE_EPAGES::Ebay::Hooks::Product::OnAfterUpdateProduct";
            $FunctionName =~ /(.*)::[^:]+$/x;
            UsePackage($1);

            foreach my $aProductItem (@$aEbayProducts){
                next unless defined $aProductItem->[0];
                my $Product = LoadObject($aProductItem->[0]);
                my $aSubProducts = $Product->get('SubProducts');
                my $aProductsToUpdate = [];

                if (scalar(@$aSubProducts)) {
                    $aProductsToUpdate = $aSubProducts;
                } else {
                    push @$aProductsToUpdate, $Product;
                }

                foreach my $ProductItem (@$aProductsToUpdate) {

                    my $ProductStocklevel = $ProductItem->get('StockLevel');
                    my $hValues= {  'Object'    => $ProductItem,
                        'ProductID'  => $ProductItem->id,
                        'IgnoreQuantityDifferences' => 1,
                        'Values' => {
                            'StockLevel' => $ProductStocklevel}
                    };
                    ## no critic (ProhibitNoStrict)
                    no strict 'refs';
                    $FunctionName->($hValues) unless $TestMode;
                    use strict 'refs';
                    ## use critic

                    my $ProductAliasLogContext = $ProductItem->alias;
                    if (defined $ProductItem->get('SuperProduct')) {
                        my $SuperProduct = $ProductItem->get('SuperProduct');
                        $ProductAliasLogContext = "Subproduct: ".$ProductAliasLogContext.">, <".$SuperProduct->alias;
                    }
                    print "<$ProductAliasLogContext>, <$ProductStocklevel> \n" unless ($NoOutput);
                }
            }
        }
    );

    my $logmsg = sprintf ('need %.1f seconds', time - $time1);
    print $logmsg."\n" unless $NoOutput;
    GetLog()->debug("Update auctions: $logmsg");

    eval { SynchronizeCache(); };
    Warning( GetError() ) if ExistsError();
}

sub _getProductWithRunningEbayAuctions {
    my ($Shop) = @_;

    # Get all products wit running ebay auctions.
    my $aProducts = GetCurrentDBHandle()->execute(<<END_SQL, [ $Shop->id ]);
                SELECT productid
                FROM ebayauction
                WHERE shopid= ?
                AND status = 2
                AND isreviseinventory = 1
                AND listingtype = 'FixedPriceItem'
                ORDER BY productid;
END_SQL
    return $aProducts;
}

sub usage {
    print <<END_USAGE;
Force new stocklevel update for running ebay auctions with type FixedPriceItem.

Usage:
    perl $0 [options] [flags]

options:
    -passwd       database user password              ()
    -storename    name of store                       (Store)
    -shop         shop alias                          ()

flags:
    -nooutput     do not print any output
    -test         Show what to do but without doing it !
    -help         show the command line options

Example:
    perl $0 -storename Store -shop DemoShop
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main
);
