#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::API::GoogleApiService
# state        public
#----------------------------------------------------------------------------------------
# description  Service to (directly) connect to Google API services
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::API::GoogleApiService;

use base DE_EPAGES::GoogleSmartShopping::API::GoogleApiClient;

use Encode qw ( encode );
use HTTP::Request;
use HTTP::Status qw ( is_server_error HTTP_OK HTTP_BAD_REQUEST );
use LWP::UserAgent;
use JSON::MaybeXS qw ( decode_json encode_json );

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Object::API::Factory qw (LoadRootObject);
use DE_EPAGES::GoogleSmartShopping::API::Constants qw (
    EPAGESJ_BASE_URI
);
use DE_EPAGES::GoogleSmartShopping::API::Tools qw ( LogGoogleSmartShopping );

use strict;

use constant GOOGLE_OPENID_USERINFO_URL => 'https://openidconnect.googleapis.com/v1/userinfo';

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new( %args );
#----------------------------------------------------------------------------------------
# description  Instantiates a new GoogleApiService.
#----------------------------------------------------------------------------------------
# return       GoogleApiService | object | DE_EPAGES::GoogleSmartShopping::API::GoogleApiService
#========================================================================================
sub new {
    my $class = shift;
    my ( $hSettings ) = @_;

    my $self = $class->SUPER::new($hSettings);

    return $self;
}


#========================================================================================
# function     getOAuthTokens
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResponse = $self->getOAuthTokens();
#----------------------------------------------------------------------------------------
# description  Given an OAuth code it gets Google API access tokens.
#----------------------------------------------------------------------------------------
# input        $OAuthCode | authorization code string | string
# return       $hResponse | hash with response | hashref
#========================================================================================
sub getOAuthTokens {
    my $self = shift;
    my ($OAuthCode) = @_;

    my $hSystemSettings = LoadRootObject()->get([qw(
        GoogleSmartShopping_AuthTokenURL
        GoogleSmartShopping_AppClientID
        GoogleSmartShopping_AppClientSecret
        GoogleSmartShopping_AppRedirectURL
    )]);

    my $Url = $hSystemSettings->{'GoogleSmartShopping_AuthTokenURL'};

    my $Header = [ 'Content-Type' => 'application/x-www-form-urlencoded' ];
    my $RequestContent = sprintf(
        'code=%s&client_id=%s&client_secret=%s&grant_type=%s&redirect_uri=%s',
        $OAuthCode,
        $hSystemSettings->{'GoogleSmartShopping_AppClientID'},
        $hSystemSettings->{'GoogleSmartShopping_AppClientSecret'},
        'authorization_code',
        $hSystemSettings->{'GoogleSmartShopping_AppRedirectURL'}
    );

    my $request = HTTP::Request->new('POST', $Url, $Header, encode("UTF-8", $RequestContent));
    my $userAgent = LWP::UserAgent->new();
    my $response = $userAgent->request($request);

    my $hResponse;
    if ( is_server_error($response->code) ) {
        $hResponse = {
            error => 'INTERNAL_SERVER_ERROR',
            errorMessage => 'Internal Server Error',
        }
    } else {
        $hResponse = decode_json($response->decoded_content());
    }

    return $hResponse;
}

#========================================================================================
# function     getUserInfo
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResponse = $self->getUserInfo($AccessToken);
#----------------------------------------------------------------------------------------
# description  Given an $AccessToken it gets email ofthe connected account
#----------------------------------------------------------------------------------------
# input        $AccessToken | access token string | string
# return       $hResponse | hash with response | hashref
#========================================================================================
sub getUserInfo {
    my $self = shift;

    my $Url = GOOGLE_OPENID_USERINFO_URL;
    my $Header = [ 'Accept'  => 'application/json' ];

    my $hUserInfo;
    eval {
        $hUserInfo = $self->doGet($Url, $Header, undef);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Error getting account user information: ', $Error->message());
    }

    return $hUserInfo;
}

#========================================================================================
# function     getUserAccountInfo
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResponse = $self->getUserAccountInfo();
#----------------------------------------------------------------------------------------
# description  Returns an OAuth code it gets Google API access tokens.
#----------------------------------------------------------------------------------------
# return       $hResponse | hash with response | hashref
#========================================================================================
sub getUserAccountInfo {
    my $self = shift;

    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';

    my $hResponse = $self->doGet($Url);

    return $hResponse;
}

#========================================================================================
# function     revokeAccessToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hResponse = $self->revokeAccessToken($AccessToken);
#----------------------------------------------------------------------------------------
# description  Revoke $AccessToken
#----------------------------------------------------------------------------------------
# input        $AccessToken | access token string | string
# return       $hResponse | hash with response | hashref
#========================================================================================
sub revokeAccessToken {
    my $self = shift;
    my ($AccessToken) = @_;

    my $Url = "https://oauth2.googleapis.com/revoke?token=$AccessToken";
    my $Header = [ 'Accept'  => 'application/json',
        'Content-type' => "application/x-www-form-urlencoded"];
    my $Data = {token => $AccessToken};
    my $Encoded_data = encode_json($Data);

    my $request = HTTP::Request->new('POST', $Url, $Header, $Encoded_data);
    my $userAgent = LWP::UserAgent->new();
    my $response = $userAgent->request($request);

    my $hResponse;
    if($response->code == HTTP_OK) {
        $hResponse = {};
    }

    if($response->code == HTTP_BAD_REQUEST) {
        my $DecodedResponse = decode_json($response->decoded_content());
        $hResponse = {
            error        => $DecodedResponse->{'error'},
            errorMessage => $DecodedResponse->{'error_description'}
        };
    } elsif ( is_server_error($response->code) ) {
        $hResponse = {
            error => 'INTERNAL_SERVER_ERROR',
            errorMessage => 'Internal Server Error',
        };
    } else {
        $hResponse = decode_json($response->decoded_content());
    }
    return $hResponse;
}

1;
