#==============================================================================
# package      DE_EPAGES::GoogleSmartShopping::UI::Shop
# state        private
#==============================================================================
package DE_EPAGES::GoogleSmartShopping::UI::Shop;
use base DE_EPAGES::Presentation::UI::Object;

use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::GoogleSmartShopping::API::GoogleApiService;
use DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder;
use DE_EPAGES::GoogleSmartShopping::API::Tools qw ( LogGoogleSmartShopping );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use HTTP::Status qw (HTTP_OK HTTP_BAD_REQUEST HTTP_FORBIDDEN HTTP_INTERNAL_SERVER_ERROR);

use strict;

use constant OAUTH_ONBOARDING => 0;
use constant ACCOUNT_CONNECTION_GET_EMAIL => 1;

#==============================================================================
# Event handlers
#==============================================================================
sub UnityViewGoogleSmartShopping {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $aSetupStages = $self->_reckonSetupStages($Shop);

    if (OAUTH_ONBOARDING ~~ $aSetupStages) {
        $Servlet->vars('GoogleOAuthConnect', $self->_buildConnectOAuthURL($Servlet, $Shop));
    }
    if (ACCOUNT_CONNECTION_GET_EMAIL ~~ $aSetupStages) {
        my $ConnectedGoogleAccountEmail = $self->_getConnectedAccountEmail($Shop);
        if (isValid($ConnectedGoogleAccountEmail)) {
            $Servlet->vars('ConnectedGoogleAccountEmail', $ConnectedGoogleAccountEmail);
        } else {
            $Servlet->vars('ConnectGoogleAccountError', 1);
        }
        $self->_setServletWithGoogleMerchantCenterRequirements($Shop, $Servlet);
    }
    return $self->SUPER::View($Servlet);
}

sub GoogleOAuthCallback {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $hQueryParameters = $Form->queryValues();
    my $payload;
    my $Code;

    if ($hQueryParameters->{'code'}) {
        LogGoogleSmartShopping('Callback Success: ', $hQueryParameters);
        my $hResponse = $self->_exchangeAuthCodeForAccessTokens($Shop, $hQueryParameters->{'code'}->[0]);
        LogGoogleSmartShopping('OAuth code exchange response: ', $hResponse);
        if ($hResponse->{'access_token'}) {
            $self->_saveTokensToConfig($Shop, $hResponse->{'access_token'}, $hResponse->{'refresh_token'});
            $payload = "window.opener.ep.fn.callbackSuccess();";
            $Code = HTTP_OK;
        } else {
            $payload = "window.opener.ep.fn.showError('" . quotemeta($hResponse->{'error'}->[0]) . "');";
            $Code = HTTP_INTERNAL_SERVER_ERROR;
        }
    } elsif ($hQueryParameters->{'error'}) {
        LogGoogleSmartShopping('Callback Error: ', $hQueryParameters);
        $payload = "window.opener.ep.fn.showError('" . quotemeta($hQueryParameters->{'error'}->[0]) . "');";
        $Code = HTTP_FORBIDDEN;
    } else {
        LogGoogleSmartShopping('Unexpected parameters to OAuth Callback: ', $hQueryParameters);
        $payload = "window.opener.ep.fn.showError('unknown_error');";
        $Code = HTTP_BAD_REQUEST;
    }
    my $jsContent = <<"EOF";
<script type="text/javascript">
    $payload
    window.close();
</script>
EOF
    return $self->_writeHTTPResponse($Servlet, $Code, {'ContentType' => 'text/html'}, $jsContent);
}

#----------------------------------------------------------------------------------------
sub _reckonSetupStages {
    my $self = shift;
    my ($Shop) = @_;

    return [ OAUTH_ONBOARDING ] unless ($Shop->existsChild('GoogleSmartShoppingConfig'));
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return [ OAUTH_ONBOARDING ] unless ($GoogleSmartShoppingConfig->get('AccessToken'));
    return [ OAUTH_ONBOARDING ] unless ($GoogleSmartShoppingConfig->get('RefreshToken'));

    my $aSecondStageMultiStep = [ ACCOUNT_CONNECTION_GET_EMAIL ];

    return $aSecondStageMultiStep;
}

#----------------------------------------------------------------------------------------
sub _buildConnectOAuthURL {
    my $self = shift;
    my ( $Servlet, $Shop ) = @_;

    my $hParams = {
        'ChangeAction' => 'GoogleOAuthCallback',
        'ObjectID'   => $Shop->id(),
    };
    my $hOptions = {
        'Type' => 'admin',
        'AddToken' => 1,
        'UseSSL' => 1,
    };
    my $CallbackUrl = BuildShopUrl($Shop, $hParams, $hOptions, $Servlet);
    LogGoogleSmartShopping('Callback URL is:', $CallbackUrl);
    my $GoogleUrlBuilder = DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder->new();

    return $GoogleUrlBuilder->buildConnectOAuthUrl($CallbackUrl);
}

#----------------------------------------------------------------------------------------
sub _exchangeAuthCodeForAccessTokens {
    my $self = shift;
    my ($Shop, $OAuthCode) = @_;

    my $GoogleService = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    return $GoogleService->getOAuthTokens($OAuthCode);
}

#----------------------------------------------------------------------------------------
sub _saveTokensToConfig {
    my $self = shift;
    my ($Shop, $AccessToken, $RefreshToken) = @_;

    unless($Shop->existsChild('GoogleSmartShoppingConfig')) {
        $Shop->insertChild('GoogleSmartShoppingConfig', { 'Alias' => 'GoogleSmartShoppingConfig' });
    }
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    $GoogleSmartShoppingConfig->set({ 'AccessToken' => $AccessToken });

    if (defined $RefreshToken) {
        $GoogleSmartShoppingConfig->set({ 'RefreshToken' => $RefreshToken });
    }

}

#----------------------------------------------------------------------------------------
sub _getConnectedAccountEmail {
    my $self = shift;
    my ($Shop) = @_;

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return unless defined $GoogleSmartShoppingConfig;

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    my $getUserInfoResult = $service->getUserInfo($GoogleSmartShoppingConfig->get('AccessToken'));
    if (defined $getUserInfoResult->{'email'}) {
        return $getUserInfoResult->{'email'};
    }

    LogGoogleSmartShopping("Google OpenID userinfo lookup failed: ", $getUserInfoResult);

    return;
}

#----------------------------------------------------------------------------------------
sub _setServletWithGoogleMerchantCenterRequirements {
    my $self = shift;
    my ($Shop, $Servlet) = @_;
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return unless(isValid($GoogleSmartShoppingConfig));

    my $hRequirements = $GoogleSmartShoppingConfig->getGoogleMerchantCenterRequirementsStatus();
    my $areRequirementsMet = 1;
    foreach my $Requirement (keys %{$hRequirements}) {
        $Servlet->vars($Requirement, $hRequirements->{$Requirement});
        $areRequirementsMet = 0 unless ($hRequirements->{$Requirement});
    }
    $Servlet->vars('GoogleMerchantCenterRequirementsMet', $areRequirementsMet);

    return;
}

sub DisconnectPersonalAccount {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    return unless defined $GoogleSmartShoppingConfig;
    my $AccessToken = $GoogleSmartShoppingConfig->get('AccessToken');

    $self->_deleteGoogleSmartShoppingConfig($GoogleSmartShoppingConfig);
    LogGoogleSmartShopping("GoogleSmartShoppingConfig deleted for Shop ", $Shop->id() );

    LogGoogleSmartShopping("Revoking access token ", $AccessToken);
    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $Shop});
    my $hRevokeTokenResult = $service->revokeAccessToken($AccessToken);
    if (defined $hRevokeTokenResult->{'error'}) {
        LogGoogleSmartShopping("Google  revoke token failed: ", $hRevokeTokenResult);
        $Servlet->form->executeFormError({
            'Name' => $hRevokeTokenResult->{'error'},
            'Reason' => $hRevokeTokenResult->{'errorMessage'}
        });
    }
}

sub _deleteGoogleSmartShoppingConfig {
    my $self = shift;
    my ($GoogleSmartShoppingConfig) = @_;

    $GoogleSmartShoppingConfig->delete();

}

#----------------------------------------------------------------------------------------
sub _writeHTTPResponse {
    my $self = shift;
    my ($Servlet, $Code, $hParams, $Content) = @_;

    $Servlet->response->code($Code);
    foreach my $key (keys %{$hParams}) {
        $Servlet->params($key, $hParams->{$key});
    }
    return $Servlet->writeContentOutput($Content);
}

1;
