use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( ExistsError GetError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw( LoadObjectByPath LoadRootObject LoadObject);
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Core::API::Warning qw (Warning);


sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $OrderObjectID;
    my $Verbose;
    my $Help;
    my $Test;

    GetOptions(
        'help'          => \$Help,
        'passwd=s'      => \$Password,
        'storename=s'   => \$StoreName,
        'shop=s'        => \$ShopAlias,
        'orderobjectid=s'      => \$OrderObjectID,
        'verbose'       => \$Verbose,
        'test'          => \$Test,
    );

    usage() if $Help;

    my $time1 = time();
    RunOnStore(
        Store => $StoreName,
        DBPassword => $Password,
        Sub => sub {
            my @Shops;

            if ( (defined $OrderObjectID) and (!defined $ShopAlias))  {
                print "\nYou need to define also 'shopalias' for the use with parameter 'orderobjectid' \n";
                usage();
            }

            #run script for specific Shop
            if (defined $ShopAlias) {
                push (@Shops,  LoadObjectByPath('/Shops/' . $ShopAlias));
            } else {
                foreach my $Shop (@{LoadObjectByPath('/Shops')->children}) {
                    push (@Shops, $Shop);
                }
            }

            eval { _CorrectShopOrders(\@Shops, $OrderObjectID, $Test); };
            Warning( GetError() ) if ExistsError();
        }
    );
    my $time2 = time;
    my $seconds = $time2-$time1;
    my $logmsg = sprintf ("\nneed %.3f seconds", $seconds);
    print $logmsg."\n";
    GetLog()->debug("restored missing TaxAreas and TaxClasses: $logmsg");

}

sub _CorrectShopOrders () {
    my ($aShops, $OrderObjectID, $Test, $Verbose) = @_;

    # Load the System object.
    my $System = LoadRootObject();
    my $SystemTaxMatrixID = $System->get('TaxMatrix')->id;

    foreach my $Shop (@$aShops) {
        #only for shop with no customized taxmatrix
        if ($Shop->get('TaxMatrix')->id ne $SystemTaxMatrixID) {
            print "\nShop: ".$Shop->alias." has customized TaxMatrix.\n";
            next;
        }
        print "\nScript running for Shop: ".$Shop->alias;

        my $aOrders = [];
        if (defined $OrderObjectID) {
            push @$aOrders, $OrderObjectID;
        } else {
            #get all corupted orders from shop
            my $sql = 'SELECT orderid FROM customerorder WHERE shopid=? AND lineitemcontainerid IN (SELECT lineitemcontainerid FROM lineitemcontainer WHERE taxareaid IS NULL)';
            # execute sql statement
            my $aaGetDBOrders = GetCurrentDBHandle()->execute($sql, [$Shop->id] );
            foreach my $OrderID (@{$aaGetDBOrders}) {
                push @$aOrders, $OrderID->[0];
            }
        }

        print "\n Found: ".scalar(@{$aOrders})." corrupt Order(s) \n";

        foreach my $OrderID (@{$aOrders}) {
            my $Order = LoadObject($OrderID);
            my $LineItemContainer = $Order->get('LineItemContainer');
            print "\n\n Updating Order: ".$Order->id if defined $Test || defined $Verbose;

            my $OrderShippingAddress = $Order->get('ShippingAddress');
            my $OrderCountryID;
            my $OrderZipcode;

            #get TaxArea from ShippingAddress or BillingAddres
            if (defined $OrderShippingAddress) {
                $OrderCountryID = $OrderShippingAddress->get('CountryID');
                $OrderZipcode = $OrderShippingAddress->get('Zipcode');
            } else {
                #use BillingAddress if we do not have a ShippingAddress
                my $OrderBillingAddress = $Order->get('BillingAddress');
                $OrderCountryID = $OrderBillingAddress->get('CountryID');
                $OrderZipcode = $OrderBillingAddress->get('Zipcode');
            }

            #get all LineItems from Order container
            my $aLineItemChildren = $Order->get('LineItemContainer')->children();
            my $LanguageID = $Order->get('LineItemContainer')->get('LanguageID');
            my $ProductTaxClass;

            foreach my $LineItem (@{$aLineItemChildren}) {
                #set taxclass for LineItemProduct
                if ($LineItem->instanceOf('LineItemProduct')) {

                    #set taxclass to taxclass of product or Superproduct
                    my $Product = $LineItem->get('Product');
                    $LineItem->set({ 'TaxClass' => $Product->get('TaxClass') }) unless defined $Test;
                    if (defined $Product->get('SuperProduct')) {
                        my $SuperProduct = $Product->get('SuperProduct');
                        $ProductTaxClass = $SuperProduct->get('TaxClass');
                    }
                    else {
                        $ProductTaxClass = $Product->get('TaxClass');
                    }
                    $LineItem->set({ 'TaxClass' => $ProductTaxClass }) unless defined $Test;
                    print "\n   Set TaxClass for LineItemProduct " . $LineItem->id . " to " . $ProductTaxClass->alias if defined $Test || defined $Verbose;
                    if (defined $Product->get('TaxClassDigital')) {
                        $LineItem->set({ 'TaxClassDigital' => $Product->get('TaxClassDigital') }) unless defined $Test;
                        print "\n   Set TaxClassDigital for LineItemProduct " . $LineItem->id . " to " . $Product->get('TaxClassDigital')->alias if defined $Test || defined $Verbose;
                    }
                }

                #set taxclass for LineItemShipping
                if ($LineItem->instanceOf('LineItemShipping')) {

                    #set taxclass to taxclass of shipping method
                    my $ShippingMethod = $LineItem->get('ShippingMethod');
                    if (defined $ShippingMethod->get('TaxClass')) {
                        $LineItem->set({ 'TaxClass' => $ShippingMethod->get('TaxClass') }) unless defined $Test;
                        print "\n   Set TaxClass for ShippingMethod " . $ShippingMethod->alias . " to " . $ShippingMethod->get('TaxClass')->alias if defined $Test || defined $Verbose;
                    }
                }

                #set taxclass LineItemTax
                if ($LineItem->instanceOf('LineItemTax')) {

                    # get taxclasses from TaxMatrix
                    my $aLineItemTaxMatrixChildren = $LineItem->get('TaxMatrix')->children();
                    foreach my $TaxClassChild (@{$aLineItemTaxMatrixChildren}) {
                        if($TaxClassChild->instanceOf('TaxClass')) {
                            if ($TaxClassChild->get('Name', $LanguageID) eq $LineItem->get('Name')) {
                                $LineItem->set({ 'TaxClass' => $TaxClassChild }) unless defined $Test;
                                print "\n   Set TaxClass for LineItemTax " . $LineItem->id . " to " . $TaxClassChild->alias if defined $Test || defined $Verbose;
                            }
                        }
                    }
                }

                #set taxclass for LineItemTaxDigital
                if ($LineItem->instanceOf('LineItemTaxDigital')) {
                    $LineItem->set({ 'TaxClassDigital' => $LineItem->get('TaxClassDigital') }) unless defined $Test;
                    print "\n   Set TaxClass for LineItemTaxDigital " . $LineItem->id . " to " . $LineItem->get('TaxClassDigital')->alias if defined $Test || defined $Verbose;
                }
            }
            my $OrderTaxArea = $Shop->get('TaxMatrix')->taxAreaByCountry($OrderCountryID, $OrderZipcode);
            $LineItemContainer->set({ 'TaxArea' => $OrderTaxArea }) unless defined $Test;
            print "\n   Set TaxArea to: " . $OrderTaxArea->alias if defined $Test || defined $Verbose;
        }
        print "\nScript finished for Shop: ".$Shop->alias."\n";
    }
    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();
}

sub usage {
    print <<END_USAGE;
Usage:
    perl $0 [options] [flags]

    Script to restore Orders which have no TaxArea after
    restoring the default TaxMatrix for the shop.

options:
    -storename     storename for the database connection (Store)
    -passwd        database user password
    -shop          Alias for a specific shop (/)
    -orderobjectid ObjectID for a specific order (/)
    -verbose       verbose output
    -test          show only details for orders to be restored, do not restore them

flags:
    -help          show the command line options
Example:
    perl $0 -storename Store -shop DemoShop -test
END_USAGE
    exit 2;
}

#planed
#Example:
#    perl $0 -path /Shops/DemoShop/Customers/1001 -test check all orders of a customer
#    perl $0 -storename MyStore -test check all orders in MyStore


RunScript(
    'Sub' => \&Main
);

1;
